from lake.models.model import Model
from lake.models.sram_model import SRAMModel
import kratos as kts


# sram wrapper model
class SRAMWrapperModel(Model):

    def __init__(self,
                 use_sram_stub,
                 sram_name,
                 data_width,
                 fw_int,
                 mem_depth,
                 mem_input_ports,
                 mem_output_ports,
                 address_width,
                 bank_num,
                 num_tiles,
                 # configuration registers passed down from top level
                 enable_chain_input,
                 enable_chain_output,
                 chain_idx_input,
                 chain_idx_output):

        # generation parameters
        self.use_sram_stub = use_sram_stub
        self.sram_name = sram_name
        self.data_width = data_width
        self.fw_int = fw_int
        self.mem_depth = mem_depth
        self.mem_input_ports = mem_input_ports
        self.mem_output_ports = mem_output_ports
        self.address_width = address_width
        self.bank_num = bank_num
        self.num_tiles = num_tiles

        # configuration registers passed down from top level
        self.enable_chain_input = enable_chain_input
        self.enable_chain_output = enable_chain_output
        self.chain_idx_input = chain_idx_input
        self.chain_idx_output = chain_idx_output

        self.chain_idx_bits = max(1, kts.clog2(num_tiles))

        self.prev_wen = 0
        self.prev_cen = 0

        self.sram = SRAMModel(data_width,
                              fw_int,
                              mem_depth,
                              num_tiles)

        # no configuration space for sram, no need to set config

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val
        # no explicit configuration space for sram wrapper module

    def interact(self, data_in, addr, cen, wen, wtsel, rtsel):

        # set chain_idx_tile
        # set mem addr
        if self.num_tiles == 1:
            chain_idx_tile = 0
            addr_sram = addr
        else:
            chain_idx_tile = addr >> (self.address_width - self.chain_idx_bits - 1)
            addr = addr & (2**(self.address_width - self.chain_idx_bits - 1) - 1)

        # set chain wen
        if (self.num_tiles == 1) or (self.enable_chain_input == 0):
            wen_chain = wen
        # enable chain input
        else:
            # write
            if wen:
                if self.chain_idx_input == chain_idx_tile:
                    wen_chain = wen
                else:
                    wen_chain = 0
            # read
            else:
                wen_chain = 0

        # set chain cen
        if (self.num_tiles == 1):
            cen_chain = cen
        # write
        elif wen:
            if self.enable_chain_input == 1:
                if self.chain_idx_input == chain_idx_tile:
                    cen_chain = cen
                else:
                    cen_chain = 0
            else:
                cen_chain = cen
        # read
        else:
            if self.enable_chain_output == 1:
                if self.chain_idx_output == chain_idx_tile:
                    cen_chain = cen
                else:
                    cen_chain = 0
            else:
                cen_chain = cen

        # registered output, watch output for combinational inputs
        # set valid data
        # read
        if not self.prev_wen:
            if self.enable_chain_output == 1:
                if self.chain_idx_output == chain_idx_tile:
                    valid_data = self.prev_cen
                else:
                    valid_data = 0
            else:
                valid_data = self.prev_cen
        # write
        else:
            valid_data = 0

        # inputs for valid data, registered output
        self.prev_wen = wen
        self.prev_cen = cen

        # cannot have self.use_sram_stub = False because we do not have an external
        # SRAM module to use...please use SRAM Stub
        assert self.use_sram_stub == 1, \
            "Do not have an external SRAM module to use...please use SRAM Stub"

        if self.use_sram_stub:
            data_out = self.sram.interact(wen=wen_chain,
                                          cen=cen_chain,
                                          addr=addr_sram,
                                          data=data_in)

        return data_out, valid_data

    def get_rd_reg(self):
        return list(self.sram.rd_reg)



from kratos import *
from lake.utils.util import increment, decrement
from lake.attributes.config_reg_attr import ConfigRegAttr


class RegFIFO(Generator):
    '''
    This module generates register-based FIFOs. These are useful
    when we only need a few entries with no prefetching needed
    '''
    def __init__(self,
                 data_width,
                 width_mult,
                 depth,
                 parallel=False,
                 break_out_rd_ptr=False):

        super().__init__(f"reg_fifo_d_{depth}_w_{width_mult}", debug=True)

        self.data_width = self.parameter("data_width", 16)
        self.data_width.value = data_width
        self.depth = depth
        self.width_mult = width_mult
        self.parallel = parallel
        self.break_out_rd_ptr = break_out_rd_ptr

        assert not (depth & (depth - 1)), "FIFO depth needs to be a power of 2"

        # CLK and RST
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")
        self._clk_en = self.input("clk_en", 1)

        # INPUTS
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=self.width_mult,
                                   explicit_array=True,
                                   packed=True)
        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=self.width_mult,
                                     explicit_array=True,
                                     packed=True)

        if self.parallel:
            self._parallel_load = self.input("parallel_load", 1)
            self._parallel_read = self.input("parallel_read", 1)
            self._num_load = self.input("num_load", clog2(self.depth) + 1)
            self._parallel_in = self.input("parallel_in",
                                           self.data_width,
                                           size=(self.depth,
                                                 self.width_mult),
                                           explicit_array=True,
                                           packed=True)

            self._parallel_out = self.output("parallel_out",
                                             self.data_width,
                                             size=(self.depth,
                                                   self.width_mult),
                                             explicit_array=True,
                                             packed=True)

        self._push = self.input("push", 1)
        self._pop = self.input("pop", 1)

        self._valid = self.output("valid", 1)

        ptr_width = max(1, clog2(self.depth))

        self._rd_ptr = self.var("rd_ptr", ptr_width)
        if self.break_out_rd_ptr:
            self._rd_ptr_out = self.output("rd_ptr_out", ptr_width)
            self.wire(self._rd_ptr_out, self._rd_ptr)
        self._wr_ptr = self.var("wr_ptr", ptr_width)
        self._read = self.var("read", 1)
        self._write = self.var("write", 1)
        self._reg_array = self.var("reg_array",
                                   self.data_width,
                                   size=(self.depth,
                                         self.width_mult),
                                   packed=True,
                                   explicit_array=True)

        self._passthru = self.var("passthru", 1)
        self._empty = self.output("empty", 1)
        self._full = self.output("full", 1)

        self._num_items = self.var("num_items", clog2(self.depth) + 1)
        # self.wire(self._full, (self._wr_ptr + 1) == self._rd_ptr)
        self.wire(self._full, self._num_items == self.depth)
        # self.wire(self._empty, self._wr_ptr == self._rd_ptr)
        self.wire(self._empty, self._num_items == 0)

        self.wire(self._read, self._pop & ~self._passthru & ~self._empty)

        self.wire(self._passthru, self._pop & self._push & self._empty)

        # Should only write

        # Boilerplate Add always @(posedge clk, ...) blocks
        if self.parallel:
            self.add_code(self.set_num_items_parallel)
            self.add_code(self.reg_array_ff_parallel)
            self.add_code(self.wr_ptr_ff_parallel)
            self.add_code(self.rd_ptr_ff_parallel)
            self.wire(self._parallel_out, self._reg_array)
            self.wire(self._write,
                      self._push & ~self._passthru & (~self._full | (self._pop | self._parallel_read)))
        else:
            self.wire(self._write, self._push & ~self._passthru & (~self._full | self._pop))
            self.add_code(self.set_num_items)
            self.add_code(self.reg_array_ff)
            self.add_code(self.wr_ptr_ff)
            self.add_code(self.rd_ptr_ff)
        self.add_code(self.data_out_ff)
        self.add_code(self.valid_ff)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def rd_ptr_ff(self):
        if ~self._rst_n:
            self._rd_ptr = 0
        elif self._read:
            self._rd_ptr = self._rd_ptr + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def rd_ptr_ff_parallel(self):
        if ~self._rst_n:
            self._rd_ptr = 0
        elif self._parallel_load | self._parallel_read:
            self._rd_ptr = 0
        elif self._read:
            self._rd_ptr = self._rd_ptr + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def wr_ptr_ff(self):
        if ~self._rst_n:
            self._wr_ptr = 0
        elif self._write:
            if self._wr_ptr == (self.depth - 1):
                self._wr_ptr = 0
            else:
                self._wr_ptr = self._wr_ptr + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def wr_ptr_ff_parallel(self):
        if ~self._rst_n:
            self._wr_ptr = 0
        elif self._parallel_load:
            self._wr_ptr = self._num_load[max(1, clog2(self.depth)) - 1, 0]
        elif self._parallel_read:
            if self._push:
                self._wr_ptr = 1
            else:
                self._wr_ptr = 0
        elif self._write:
            self._wr_ptr = self._wr_ptr + 1
            # if self._wr_ptr == (self.depth - 1):
            #     self._wr_ptr = 0
            # else:
            #     self._wr_ptr = self._wr_ptr + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def reg_array_ff(self):
        if ~self._rst_n:
            self._reg_array = 0
        elif self._write:
            self._reg_array[self._wr_ptr] = self._data_in

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def reg_array_ff_parallel(self):
        if ~self._rst_n:
            self._reg_array = 0
        elif self._parallel_load:
            self._reg_array = self._parallel_in
        elif self._write:
            if self._parallel_read:
                self._reg_array[0] = self._data_in
            else:
                self._reg_array[self._wr_ptr] = self._data_in

    @always_comb
    def data_out_ff(self):
        if(self._passthru):
            self._data_out = self._data_in
        else:
            self._data_out = self._reg_array[self._rd_ptr]

    @always_comb
    def valid_ff(self):
        self._valid = self._pop & ((~self._empty) | self._passthru)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_num_items(self):
        if ~self._rst_n:
            self._num_items = 0
        elif self._write & ~self._read:
            self._num_items = self._num_items + 1
        elif ~self._write & self._read:
            self._num_items = self._num_items - 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_num_items_parallel(self):
        if ~self._rst_n:
            self._num_items = 0
        elif self._parallel_load:
            # When fetch width > 1, by definition we cannot load
            # 0 (only fw or fw - 1), but we need to handle this immediate
            # pass through in these cases
            if self._num_load == 0:
                self._num_items = self._push.extend(self._num_items.width)
            else:
                self._num_items = self._num_load
        # One can technically push while a parallel
        # read is happening...
        elif self._parallel_read:
            if self._push:
                self._num_items = 1
            else:
                self._num_items = 0
        elif self._write & ~self._read:
            self._num_items = self._num_items + 1
        elif ~self._write & self._read:
            self._num_items = self._num_items - 1


if __name__ == "__main__":
    dut = RegFIFO(data_width=16,
                  depth=64,
                  width_mult=4)
    verilog(dut, filename="reg_fifo.sv")



from kratos import *
from lake.modules.passthru import *
from lake.modules.aggregation_buffer import AggregationBuffer
from lake.modules.input_addr_ctrl import InputAddrCtrl
from lake.modules.output_addr_ctrl import OutputAddrCtrl
from lake.modules.agg_aligner import AggAligner
from lake.modules.rw_arbiter import RWArbiter
from lake.modules.transpose_buffer import TransposeBuffer
from lake.modules.transpose_buffer_aggregation import TransposeBufferAggregation
from lake.modules.demux_reads import DemuxReads
from lake.modules.sync_groups import SyncGroups
from lake.modules.prefetcher import Prefetcher
from lake.modules.register_file import RegisterFile
from lake.modules.strg_fifo import StrgFIFO
from lake.modules.strg_RAM import StrgRAM
from lake.modules.app_ctrl import AppCtrl
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
import kratos as kts


class StrgUB(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_iterator_support=6,  # Addr Controllers
                 output_iterator_support=6,
                 input_config_width=16,
                 output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 max_agg_schedule=32,
                 input_max_port_sched=32,
                 output_max_port_sched=32,
                 align_input=1,
                 max_line_length=128,
                 max_tb_height=1,
                 tb_range_max=128,
                 tb_range_inner_max=5,
                 tb_sched_max=64,
                 max_tb_stride=15,
                 num_tb=1,
                 tb_iterator_support=2,
                 multiwrite=1,
                 num_tiles=1,
                 max_prefetch=8,
                 app_ctrl_depth_width=16,
                 remove_tb=False,
                 stcl_valid_iter=4):
        super().__init__("strg_ub")

        self.data_width = data_width
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.banks = banks
        self.input_iterator_support = input_iterator_support
        self.output_iterator_support = output_iterator_support
        self.input_config_width = input_config_width
        self.output_config_width = output_config_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.mem_input_ports = mem_input_ports
        self.mem_output_ports = mem_output_ports
        self.agg_height = agg_height
        self.max_agg_schedule = max_agg_schedule
        self.input_max_port_sched = input_max_port_sched
        self.output_max_port_sched = output_max_port_sched
        self.input_port_sched_width = clog2(self.interconnect_input_ports)
        self.align_input = align_input
        self.max_line_length = max_line_length
        assert self.mem_width >= self.data_width, "Data width needs to be smaller than mem"
        self.fw_int = int(self.mem_width / self.data_width)
        self.num_tb = num_tb
        self.max_tb_height = max_tb_height
        self.tb_range_max = tb_range_max
        self.tb_range_inner_max = tb_range_inner_max
        self.max_tb_stride = max_tb_stride
        self.tb_sched_max = tb_sched_max
        self.tb_iterator_support = tb_iterator_support
        self.multiwrite = multiwrite
        self.max_prefetch = max_prefetch
        self.num_tiles = num_tiles
        self.app_ctrl_depth_width = app_ctrl_depth_width
        self.remove_tb = remove_tb
        self.read_delay = read_delay
        self.rw_same_cycle = rw_same_cycle
        self.stcl_valid_iter = stcl_valid_iter
        # phases = [] TODO

        self.address_width = clog2(self.num_tiles * self.mem_depth)

        # CLK and RST
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # INPUTS
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        self._wen_in = self.input("wen_in", self.interconnect_input_ports)
        self._ren_input = self.input("ren_in", self.interconnect_output_ports)
        # Post rate matched
        self._ren_in = self.var("ren_in_muxed", self.interconnect_output_ports)
        # Processed versions of wen and ren from the app ctrl
        self._wen = self.var("wen", self.interconnect_input_ports)
        self._ren = self.var("ren", self.interconnect_output_ports)

        # Add rate matched
        # If one input port, let any output port use the wen_in as the ren_in
        # If more, do the same thing but also provide port selection
        if self.interconnect_input_ports == 1:
            self._rate_matched = self.input("rate_matched", self.interconnect_output_ports)
            self._rate_matched.add_attribute(ConfigRegAttr("Rate matched - 1 or 0"))
            for i in range(self.interconnect_output_ports):
                self.wire(self._ren_in[i],
                          kts.ternary(self._rate_matched[i],
                                      self._wen_in,
                                      self._ren_input[i]))
        else:
            self._rate_matched = self.input("rate_matched", 1 + kts.clog2(self.interconnect_input_ports),
                                            size=self.interconnect_output_ports,
                                            explicit_array=True,
                                            packed=True)
            self._rate_matched.add_attribute(ConfigRegAttr("Rate matched [input port | on/off]"))
            for i in range(self.interconnect_output_ports):
                self.wire(self._ren_in[i],
                          kts.ternary(self._rate_matched[i][0],
                                      self._wen_in[self._rate_matched[i][kts.clog2(self.interconnect_input_ports), 1]],
                                      self._ren_input[i]))

        self._arb_wen_en = self.var("arb_wen_en", self.interconnect_input_ports)
        self._arb_ren_en = self.var("arb_ren_en", self.interconnect_output_ports)

        self._data_from_strg = self.input("data_from_strg",
                                          self.data_width,
                                          size=(self.banks,
                                                self.mem_output_ports,
                                                self.fw_int),
                                          packed=True,
                                          explicit_array=True)

        self._mem_valid_data = self.input("mem_valid_data",
                                          self.mem_output_ports,
                                          size=self.banks,
                                          explicit_array=True,
                                          packed=True)

        self._out_mem_valid_data = self.var("out_mem_valid_data",
                                            self.mem_output_ports,
                                            size=self.banks,
                                            explicit_array=True,
                                            packed=True)

        # We need to signal valids out of the agg buff, only if one exists...
        if self.agg_height > 0:
            self._to_iac_valid = self.var("ab_to_mem_valid",
                                          self.interconnect_input_ports)

        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        self._valid_out = self.output("valid_out",
                                      self.interconnect_output_ports)

        self._valid_out_alt = self.var("valid_out_alt",
                                       self.interconnect_output_ports)

        self._data_to_strg = self.output("data_to_strg",
                                         self.data_width,
                                         size=(self.banks,
                                               self.mem_input_ports,
                                               self.fw_int),
                                         packed=True,
                                         explicit_array=True)

        # If we can perform a read and a write on the same cycle,
        # this will necessitate a separate read and write address...
        if self.rw_same_cycle:
            self._wr_addr_out = self.output("wr_addr_out",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_input_ports),
                                            explicit_array=True,
                                            packed=True)

            self._rd_addr_out = self.output("rd_addr_out",
                                            self.address_width,
                                            size=(self.banks,
                                                  self.mem_output_ports),
                                            explicit_array=True,
                                            packed=True)

        else:
            self._addr_out = self.output("addr_out",
                                         self.address_width,
                                         size=(self.banks,
                                               self.mem_input_ports),
                                         packed=True,
                                         explicit_array=True)

        self._cen_to_strg = self.output("cen_to_strg", self.mem_output_ports,
                                        size=self.banks,
                                        explicit_array=True,
                                        packed=True)
        self._wen_to_strg = self.output("wen_to_strg", self.mem_input_ports,
                                        size=self.banks,
                                        explicit_array=True,
                                        packed=True)
        if self.num_tb > 0:
            self._tb_valid_out = self.var("tb_valid_out", self.interconnect_output_ports)

        self._port_wens = self.var("port_wens", self.interconnect_input_ports)

        ####################
        ##### APP CTRL #####
        ####################
        self._ack_transpose = self.var("ack_transpose",
                                       self.banks,
                                       size=self.interconnect_output_ports,
                                       explicit_array=True,
                                       packed=True)

        self._ack_reduced = self.var("ack_reduced",
                                     self.interconnect_output_ports)

        self.app_ctrl = AppCtrl(interconnect_input_ports=self.interconnect_input_ports,
                                interconnect_output_ports=self.interconnect_output_ports,
                                depth_width=self.app_ctrl_depth_width,
                                sprt_stcl_valid=True,
                                stcl_iter_support=self.stcl_valid_iter)

        # Some refactoring here for pond to get rid of app controllers...
        # This is honestly pretty messy and should clean up nicely when we have the spec...
        self._ren_out_reduced = self.var("ren_out_reduced",
                                         self.interconnect_output_ports)

        if self.num_tb == 0 or self.remove_tb:
            self.wire(self._wen, self._wen_in)
            self.wire(self._ren, self._ren_in)
            self.wire(self._valid_out, self._valid_out_alt)
            self.wire(self._arb_wen_en, self._wen)
            self.wire(self._arb_ren_en, self._ren)
        else:
            self.add_child("app_ctrl", self.app_ctrl,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           wen_in=self._wen_in,
                           ren_in=self._ren_in,
                           #    ren_update=self._tb_valid_out,
                           valid_out_data=self._valid_out,
                           # valid_out_stencil=,
                           wen_out=self._wen,
                           ren_out=self._ren)

            self.wire(self.app_ctrl.ports.tb_valid, self._tb_valid_out)
            self.wire(self.app_ctrl.ports.ren_update, self._tb_valid_out)

            self.app_ctrl_coarse = AppCtrl(interconnect_input_ports=self.interconnect_input_ports,
                                           interconnect_output_ports=self.interconnect_output_ports,
                                           depth_width=self.app_ctrl_depth_width)
            self.add_child("app_ctrl_coarse", self.app_ctrl_coarse,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           wen_in=self._to_iac_valid,  # self._port_wens & self._to_iac_valid,  # Gets valid and the ack
                           ren_in=self._ren_out_reduced,
                           tb_valid=kts.const(0, 1),
                           ren_update=self._ack_reduced,
                           wen_out=self._arb_wen_en,
                           ren_out=self._arb_ren_en)

        ###########################
        ##### INPUT AGG SCHED #####
        ###########################
        ###########################################
        ##### AGGREGATION ALIGNERS (OPTIONAL) #####
        ###########################################
        # These variables are holders and can be swapped out if needed
        self._data_consume = self._data_in
        self._valid_consume = self._wen
        # Zero out if not aligning
        if(self.agg_height > 0):
            self._align_to_agg = self.var("align_input",
                                          self.interconnect_input_ports)
        # Add the aggregation buffer aligners
        if(self.align_input):
            self._data_consume = self.var("data_consume",
                                          self.data_width,
                                          size=self.interconnect_input_ports,
                                          packed=True,
                                          explicit_array=True)
            self._valid_consume = self.var("valid_consume",
                                           self.interconnect_input_ports)

            # Make new aggregation aligners for each port
            for i in range(self.interconnect_input_ports):
                new_child = AggAligner(self.data_width, self.max_line_length)
                self.add_child(f"agg_align_{i}", new_child,
                               clk=self._clk,
                               rst_n=self._rst_n,
                               in_dat=self._data_in[i],
                               in_valid=self._wen[i],
                               align=self._align_to_agg[i],
                               out_valid=self._valid_consume[i],
                               out_dat=self._data_consume[i])
        else:
            if self.agg_height > 0:
                self.wire(self._align_to_agg, const(0, self._align_to_agg.width))
        ################################################
        ##### END: AGGREGATION ALIGNERS (OPTIONAL) #####
        ################################################

        if self.agg_height == 0:
            self._to_iac_dat = self._data_consume
            self._to_iac_valid = self._valid_consume

        ##################################
        ##### AGG BUFFERS (OPTIONAL) #####
        ##################################
        # Only instantiate agg_buffer if needed
        if(self.agg_height > 0):
            self._to_iac_dat = self.var("ab_to_mem_dat",
                                        self.mem_width,
                                        size=self.interconnect_input_ports,
                                        packed=True,
                                        explicit_array=True)

            # self._to_iac_valid = self.var("ab_to_mem_valid",
            #                               self.interconnect_input_ports)

            self._agg_buffers = []
            # Add input aggregations buffers
            for i in range(self.interconnect_input_ports):
                # add children aggregator buffers...
                agg_buffer_new = AggregationBuffer(self.agg_height,
                                                   self.data_width,
                                                   self.mem_width,
                                                   self.max_agg_schedule)
                self._agg_buffers.append(agg_buffer_new)
                self.add_child(f"agg_in_{i}", agg_buffer_new,
                               clk=self._clk,
                               rst_n=self._rst_n,
                               data_in=self._data_consume[i],
                               valid_in=self._valid_consume[i],
                               align=self._align_to_agg[i],
                               data_out=self._to_iac_dat[i],
                               valid_out=self._to_iac_valid[i])

        #######################################
        ##### END: AGG BUFFERS (OPTIONAL) #####
        #######################################

        self._ready_tba = self.var("ready_tba", self.interconnect_output_ports)
        ####################################
        ##### INPUT ADDRESS CONTROLLER #####
        ####################################
        self._wen_to_arb = self.var("wen_to_arb", self.mem_input_ports,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)
        self._addr_to_arb = self.var("addr_to_arb",
                                     self.address_width,
                                     size=(self.banks,
                                           self.mem_input_ports),
                                     explicit_array=True,
                                     packed=True)
        self._data_to_arb = self.var("data_to_arb",
                                     self.data_width,
                                     size=(self.banks,
                                           self.mem_input_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        # Connect these inputs ports to an address generator
        iac = InputAddrCtrl(interconnect_input_ports=self.interconnect_input_ports,
                            mem_depth=self.mem_depth,
                            num_tiles=self.num_tiles,
                            banks=self.banks,
                            iterator_support=self.input_iterator_support,
                            address_width=self.address_width,
                            data_width=self.data_width,
                            fetch_width=self.mem_width,
                            multiwrite=self.multiwrite,
                            strg_wr_ports=self.mem_input_ports,
                            config_width=self.input_config_width)
        self.add_child(f"input_addr_ctrl", iac,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       valid_in=self._to_iac_valid,
                       # wen_en=kts.concat(*([kts.const(1, 1)] * self.interconnect_input_ports)),
                       wen_en=self._arb_wen_en,
                       data_in=self._to_iac_dat,
                       wen_to_sram=self._wen_to_arb,
                       addr_out=self._addr_to_arb,
                       port_out=self._port_wens,
                       data_out=self._data_to_arb)

        #########################################
        ##### END: INPUT ADDRESS CONTROLLER #####
        #########################################
        self._arb_acks = self.var("arb_acks",
                                  self.interconnect_output_ports,
                                  size=self.banks,
                                  explicit_array=True,
                                  packed=True)

        self._prefetch_step = self.var("prefetch_step", self.interconnect_output_ports)
        self._oac_step = self.var("oac_step", self.interconnect_output_ports)
        self._oac_valid = self.var("oac_valid", self.interconnect_output_ports)
        self._ren_out = self.var("ren_out",
                                 self.interconnect_output_ports,
                                 size=self.banks,
                                 explicit_array=True,
                                 packed=True)
        self._ren_out_tpose = self.var("ren_out_tpose",
                                       self.banks,
                                       size=self.interconnect_output_ports,
                                       explicit_array=True,
                                       packed=True)

        self._oac_addr_out = self.var("oac_addr_out",
                                      self.address_width,
                                      size=self.interconnect_output_ports,
                                      explicit_array=True,
                                      packed=True)
        #####################################
        ##### OUTPUT ADDRESS CONTROLLER #####
        #####################################
        oac = OutputAddrCtrl(interconnect_output_ports=self.interconnect_output_ports,
                             mem_depth=self.mem_depth,
                             num_tiles=self.num_tiles,
                             banks=self.banks,
                             iterator_support=self.output_iterator_support,
                             address_width=self.address_width,
                             config_width=self.output_config_width)

        if self.remove_tb:
            self.wire(self._oac_valid, self._ren)
            self.wire(self._oac_step, self._ren)
        else:
            self.wire(self._oac_valid, self._prefetch_step)
            self.wire(self._oac_step, self._ack_reduced)

        self.chain_idx_bits = max(1, clog2(num_tiles))
        self._enable_chain_output = self.input("enable_chain_output", 1)
        self._chain_idx_output = self.input("chain_idx_output", self.chain_idx_bits)

        self.add_child(f"output_addr_ctrl", oac,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       valid_in=self._oac_valid,
                       ren=self._ren_out,
                       addr_out=self._oac_addr_out,
                       step_in=self._oac_step)

        for i in range(self.interconnect_output_ports):
            for j in range(self.banks):
                self.wire(self._ren_out_tpose[i][j], self._ren_out[j][i])
        ##############################
        ##### READ/WRITE ARBITER #####
        ##############################
        # Hook up the read write arbiters for each bank
        self._arb_dat_out = self.var("arb_dat_out",
                                     self.data_width,
                                     size=(self.banks,
                                           self.mem_output_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        self._arb_port_out = self.var("arb_port_out",
                                      self.interconnect_output_ports,
                                      size=(self.banks,
                                            self.mem_output_ports),
                                      explicit_array=True,
                                      packed=True)
        self._arb_valid_out = self.var("arb_valid_out", self.mem_output_ports,
                                       size=self.banks,
                                       explicit_array=True,
                                       packed=True)

        self._rd_sync_gate = self.var("rd_sync_gate",
                                      self.interconnect_output_ports)

        self.arbiters = []
        for i in range(self.banks):
            rw_arb = RWArbiter(fetch_width=self.mem_width,
                               data_width=self.data_width,
                               memory_depth=self.mem_depth,
                               num_tiles=self.num_tiles,
                               int_in_ports=self.interconnect_input_ports,
                               int_out_ports=self.interconnect_output_ports,
                               strg_wr_ports=self.mem_input_ports,
                               strg_rd_ports=self.mem_output_ports,
                               read_delay=self.read_delay,
                               rw_same_cycle=self.rw_same_cycle,
                               separate_addresses=self.rw_same_cycle)
            self.arbiters.append(rw_arb)
            self.add_child(f"rw_arb_{i}", rw_arb,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           wen_in=self._wen_to_arb[i],
                           w_data=self._data_to_arb[i],
                           w_addr=self._addr_to_arb[i],
                           data_from_mem=self._data_from_strg[i],
                           mem_valid_data=self._mem_valid_data[i],
                           out_mem_valid_data=self._out_mem_valid_data[i],
                           ren_en=self._arb_ren_en,
                           rd_addr=self._oac_addr_out,
                           out_data=self._arb_dat_out[i],
                           out_port=self._arb_port_out[i],
                           out_valid=self._arb_valid_out[i],
                           cen_mem=self._cen_to_strg[i],
                           wen_mem=self._wen_to_strg[i],
                           data_to_mem=self._data_to_strg[i],
                           out_ack=self._arb_acks[i])

            # Bind the separate addrs
            if self.rw_same_cycle:
                self.wire(rw_arb.ports.wr_addr_to_mem, self._wr_addr_out[i])
                self.wire(rw_arb.ports.rd_addr_to_mem, self._rd_addr_out[i])
            else:
                self.wire(rw_arb.ports.addr_to_mem, self._addr_out[i])

            if self.remove_tb:
                self.wire(rw_arb.ports.ren_in, self._ren_out[i])
            else:
                self.wire(rw_arb.ports.ren_in, self._ren_out[i] & self._rd_sync_gate)

        self.num_tb_bits = max(1, clog2(self.num_tb))

        self._data_to_sync = self.var("data_to_sync",
                                      self.data_width,
                                      size=(self.interconnect_output_ports,
                                            self.fw_int),
                                      explicit_array=True,
                                      packed=True)

        self._valid_to_sync = self.var("valid_to_sync", self.interconnect_output_ports)

        self._data_to_tba = self.var("data_to_tba",
                                     self.data_width,
                                     size=(self.interconnect_output_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        self._valid_to_tba = self.var("valid_to_tba", self.interconnect_output_ports)

        self._data_to_pref = self.var("data_to_pref",
                                      self.data_width,
                                      size=(self.interconnect_output_ports,
                                            self.fw_int),
                                      explicit_array=True,
                                      packed=True)

        self._valid_to_pref = self.var("valid_to_pref", self.interconnect_output_ports)
        #######################
        ##### DEMUX READS #####
        #######################
        dmux_rd = DemuxReads(fetch_width=self.mem_width,
                             data_width=self.data_width,
                             banks=self.banks,
                             int_out_ports=self.interconnect_output_ports,
                             strg_rd_ports=self.mem_output_ports)

        self._arb_dat_out_f = self.var("arb_dat_out_f",
                                       self.data_width,
                                       size=(self.banks * self.mem_output_ports,
                                             self.fw_int),
                                       explicit_array=True,
                                       packed=True)

        self._arb_port_out_f = self.var("arb_port_out_f",
                                        self.interconnect_output_ports,
                                        size=(self.banks * self.mem_output_ports),
                                        explicit_array=True,
                                        packed=True)
        self._arb_valid_out_f = self.var("arb_valid_out_f", self.mem_output_ports * self.banks)
        self._arb_mem_valid_data_f = self.var("arb_mem_valid_data_f", self.mem_output_ports * self.banks)

        self._arb_mem_valid_data_out = self.var("arb_mem_valid_data_out",
                                                self.interconnect_output_ports)

        self._mem_valid_data_sync = self.var("mem_valid_data_sync",
                                             self.interconnect_output_ports)

        self._mem_valid_data_pref = self.var("mem_valid_data_pref",
                                             self.interconnect_output_ports)

        tmp_cnt = 0
        for i in range(self.banks):
            for j in range(self.mem_output_ports):
                self.wire(self._arb_dat_out_f[tmp_cnt], self._arb_dat_out[i][j])
                self.wire(self._arb_port_out_f[tmp_cnt], self._arb_port_out[i][j])
                self.wire(self._arb_valid_out_f[tmp_cnt], self._arb_valid_out[i][j])
                self.wire(self._arb_mem_valid_data_f[tmp_cnt], self._out_mem_valid_data[i][j])
                tmp_cnt = tmp_cnt + 1

        # If this is end of the road...
        if self.remove_tb:
            assert self.fw_int == 1, "Make it easier on me now..."
            self.add_child("demux_rds", dmux_rd,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           data_in=self._arb_dat_out_f,
                           mem_valid_data=self._arb_mem_valid_data_f,
                           mem_valid_data_out=self._arb_mem_valid_data_out,
                           valid_in=self._arb_valid_out_f,
                           port_in=self._arb_port_out_f,
                           valid_out=self._valid_out_alt)
            for i in range(self.interconnect_output_ports):
                self.wire(self._data_out[i], dmux_rd.ports.data_out[i])

        else:
            self.add_child("demux_rds", dmux_rd,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           data_in=self._arb_dat_out_f,
                           mem_valid_data=self._arb_mem_valid_data_f,
                           mem_valid_data_out=self._arb_mem_valid_data_out,
                           valid_in=self._arb_valid_out_f,
                           port_in=self._arb_port_out_f,
                           data_out=self._data_to_sync,
                           valid_out=self._valid_to_sync)

            #######################
            ##### SYNC GROUPS #####
            #######################
            sync_group = SyncGroups(fetch_width=self.mem_width,
                                    data_width=self.data_width,
                                    int_out_ports=self.interconnect_output_ports)

            for i in range(self.interconnect_output_ports):
                self.wire(self._ren_out_reduced[i], self._ren_out_tpose[i].r_or())

            self.add_child("sync_grp", sync_group,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           data_in=self._data_to_sync,
                           mem_valid_data=self._arb_mem_valid_data_out,
                           mem_valid_data_out=self._mem_valid_data_sync,
                           valid_in=self._valid_to_sync,
                           data_out=self._data_to_pref,
                           valid_out=self._valid_to_pref,
                           ren_in=self._ren_out_reduced,
                           rd_sync_gate=self._rd_sync_gate,
                           ack_in=self._ack_reduced)

            # This is the end of the line if we aren't using tb
            ######################
            ##### PREFETCHER #####
            ######################
            prefetchers = []
            for i in range(self.interconnect_output_ports):

                pref = Prefetcher(fetch_width=self.mem_width,
                                  data_width=self.data_width,
                                  max_prefetch=self.max_prefetch)

                prefetchers.append(pref)

                if self.num_tb == 0:
                    assert self.fw_int == 1, \
                        "If no transpose buffer, data width needs match memory width"
                    self.add_child(f"pre_fetch_{i}", pref,
                                   clk=self._clk,
                                   rst_n=self._rst_n,
                                   data_in=self._data_to_pref[i],
                                   mem_valid_data=self._mem_valid_data_sync[i],
                                   mem_valid_data_out=self._mem_valid_data_pref[i],
                                   valid_read=self._valid_to_pref[i],
                                   tba_rdy_in=self._ren[i],
                                   #    data_out=self._data_out[i],
                                   valid_out=self._valid_out_alt[i],
                                   prefetch_step=self._prefetch_step[i])
                    self.wire(self._data_out[i], pref.ports.data_out[0])
                else:
                    self.add_child(f"pre_fetch_{i}", pref,
                                   clk=self._clk,
                                   rst_n=self._rst_n,
                                   data_in=self._data_to_pref[i],
                                   mem_valid_data=self._mem_valid_data_sync[i],
                                   mem_valid_data_out=self._mem_valid_data_pref[i],
                                   valid_read=self._valid_to_pref[i],
                                   tba_rdy_in=self._ready_tba[i],
                                   data_out=self._data_to_tba[i],
                                   valid_out=self._valid_to_tba[i],
                                   prefetch_step=self._prefetch_step[i])

                    #############################
                    ##### TRANSPOSE BUFFERS #####
                    #############################
            if self.num_tb > 0:

                self._tb_data_out = self.var("tb_data_out", self.data_width,
                                             size=self.interconnect_output_ports,
                                             explicit_array=True,
                                             packed=True)
                self._tb_valid_out = self.var("tb_valid_out", self.interconnect_output_ports)

                for i in range(self.interconnect_output_ports):

                    tba = TransposeBufferAggregation(word_width=self.data_width,
                                                     fetch_width=self.fw_int,
                                                     num_tb=self.num_tb,
                                                     max_tb_height=self.max_tb_height,
                                                     max_range=self.tb_range_max,
                                                     max_range_inner=self.tb_range_inner_max,
                                                     max_stride=self.max_tb_stride,
                                                     tb_iterator_support=self.tb_iterator_support)

                    self.add_child(f"tba_{i}", tba,
                                   clk=self._clk,
                                   rst_n=self._rst_n,
                                   SRAM_to_tb_data=self._data_to_tba[i],
                                   valid_data=self._valid_to_tba[i],
                                   tb_index_for_data=0,
                                   ack_in=self._valid_to_tba[i],
                                   mem_valid_data=self._mem_valid_data_pref[i],
                                   tb_to_interconnect_data=self._tb_data_out[i],
                                   tb_to_interconnect_valid=self._tb_valid_out[i],
                                   tb_arbiter_rdy=self._ready_tba[i],
                                   tba_ren=self._ren[i])

                for i in range(self.interconnect_output_ports):
                    self.wire(self._data_out[i], self._tb_data_out[i])
                    # self.wire(self._valid_out[i], self._tb_valid_out[i])
            else:
                self.wire(self._valid_out, self._valid_out_alt)

        ####################
        ##### ADD CODE #####
        ####################
        self.add_code(self.transpose_acks)
        self.add_code(self.reduce_acks)

    @always_comb
    def transpose_acks(self):
        for i in range(self.interconnect_output_ports):
            for j in range(self.banks):
                self._ack_transpose[i][j] = self._arb_acks[j][i]

    @always_comb
    def reduce_acks(self):
        for i in range(self.interconnect_output_ports):
            self._ack_reduced[i] = self._ack_transpose[i].r_or()


if __name__ == "__main__":
    lake_dut = StrgUB()
    verilog(lake_dut, filename="lake_top.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from lake.models.model import Model


class AggModel(Model):

    def __init__(self,
                 num_elts):

        self.num_elts = num_elts

        self.shift_reg = []
        for i in range(self.num_elts):
            self.shift_reg.append(0)

        self.word_count = 0
        self.valid_out = 0
        self.curr_valid = 0
        self.sr_copy = 0

    def set_config(self, **kwargs):
        self.word_count = 0
        self.valid_out = 0
        return

    def get_valid_out(self):
        return self.curr_valid

    def get_data_out(self):
        return self.sr_copy

    def interact(self, data, valid_in, align):
        '''
            Returns (agg_out, valid_out, next_full)
        '''

        self.curr_valid = self.valid_out
        self.sr_copy = self.shift_reg.copy()
        word_count_curr = self.word_count

        next_full = (valid_in & (word_count_curr == self.num_elts - 1)) | align
        if valid_in:
            if next_full:
                self.valid_out = 1
                self.word_count = 0
            else:
                self.valid_out = 0
                self.word_count = word_count_curr + 1
            self.shift_reg[word_count_curr] = data
        else:
            self.valid_out = 0
        return (self.sr_copy, self.curr_valid, next_full)



from lake.models.model import Model
from lake.models.agg_model import AggModel


class AggBuffModel(Model):

    def __init__(self,
                 agg_height,
                 data_width,
                 mem_width,
                 max_agg_schedule):

        self.agg_height = agg_height
        self.data_width = data_width
        self.mem_width = mem_width
        self.max_agg_sched = max_agg_schedule

        self.config = {}
        self.config['in_period'] = 0
        self.config['out_period'] = 0

        self.in_sched_ptr = 0
        self.out_sched_ptr = 0

        for i in range(self.max_agg_sched):
            self.config[f"in_sched_{i}"] = 0
            self.config[f"out_sched_{i}"] = 0

        self.aggs = []
        for i in range(self.agg_height):
            self.aggs.append(AggModel(int(self.mem_width / self.data_width)))
            self.aggs[i].set_config()

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val
        self.in_sched_ptr = 0
        self.out_sched_ptr = 0

    def insert(self, in_data, valid):
        to_insert = self.aggs[self.config[f"in_sched_{self.in_sched_ptr}"]]
        ag_valid = to_insert.insert(in_data, valid)

        if(ag_valid):
            self.in_sched_ptr += 1
            if(self.in_sched_ptr >= self.config['in_period']):
                self.in_sched_ptr = 0

    def get_valid_out(self):
        valid_check_agg = self.aggs[self.config[f"out_sched_{self.out_sched_ptr}"]]
        return valid_check_agg.get_valid_out()

    def get_item(self):
        valid_check_agg = self.aggs[self.config[f"out_sched_{self.out_sched_ptr}"]]
        self.out_sched_ptr += 1
        if(self.out_sched_ptr >= self.config['out_period']):
            self.out_sched_ptr = 0
        return valid_check_agg.get_data_out()

    def interact(self, in_data, valid, write_act, align):
        '''
            Returns (data_out, valid_out)
        '''
        in_sched_curr = self.in_sched_ptr
        out_sched_curr = self.out_sched_ptr

        agg_dat = []
        agg_valid = []
        agg_nf = []
        for i in range(self.agg_height):
            # Interact with child
            if i == self.config[f"in_sched_{in_sched_curr}"]:
                (t_ag_d, t_ag_v, t_nf) = self.aggs[i].interact(in_data, valid, align)
            else:
                (t_ag_d, t_ag_v, t_nf) = self.aggs[i].interact(0, 0, 0)
            agg_dat.append(t_ag_d)
            agg_valid.append(t_ag_v)
            agg_nf.append(t_nf)

        ret_data = agg_dat[self.config[f"out_sched_{out_sched_curr}"]]
        ret_valid = agg_valid[self.config[f"out_sched_{out_sched_curr}"]]

        next_full = agg_nf[self.config[f"in_sched_{in_sched_curr}"]]

        if(next_full == 1):
            self.in_sched_ptr += 1
            if(self.in_sched_ptr >= self.config['in_period']):
                self.in_sched_ptr = 0

        # Would be write_act
        if(ret_valid == 1):
            self.out_sched_ptr += 1
            if(self.out_sched_ptr >= self.config['out_period']):
                self.out_sched_ptr = 0

        return (ret_data, ret_valid)

    def peek(self):
        pass



from lake.models.model import Model


class AddrGenModel(Model):

    def __init__(self, iterator_support, address_width):
        self.iterator_support = iterator_support
        self.address_width = address_width

        self.config = {}

        self.config["starting_addr"] = 0
        self.config["dimensionality"] = 0

        self.dim_cnt = []

        self.address = 0

        for i in range(self.iterator_support):
            self.config[f"ranges_{i}"] = 0
            self.config[f"strides_{i}"] = 0
            self.dim_cnt.append(0)

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val
        for i in range(self.iterator_support):
            self.dim_cnt[i] = 0
        self.address = 0 + self.config["starting_addr"]

    def get_address(self):
        return self.address

    def step(self):
        for i in range(self.config["dimensionality"]):
            if(i == 0):
                update_curr = True

            if update_curr:
                self.dim_cnt[i] = self.dim_cnt[i] + 1
                if(self.dim_cnt[i] == self.config[f"ranges_{i}"]):
                    self.dim_cnt[i] = 0
                else:
                    break
            else:
                break

        self.address = self.config["starting_addr"]
        for i in range(self.config["dimensionality"]):
            offset = self.dim_cnt[i] * self.config[f"strides_{i}"]
            self.address = self.address + offset



from kratos import *
from lake.modules.aggregator import Aggregator
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
import kratos as kts


class StorageConfigSeq(Generator):
    '''
    Sequence the reads and writes to the storage unit - if dealing with
    a storage unit that has multiple r/w, should only use one of the ports

    If the storage unit is wider than the data, this sequencer expects all
    X writes/reads in order will come from the same word
    '''
    def __init__(self,
                 data_width,
                 config_addr_width,
                 addr_width,
                 fetch_width,
                 total_sets,
                 sets_per_macro):
        super().__init__("storage_config_seq")

        self.data_width = data_width
        self.config_addr_width = config_addr_width
        self.addr_width = addr_width
        self.fetch_width = fetch_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.total_sets = total_sets
        self.sets_per_macro = sets_per_macro
        self.banks = int(self.total_sets / self.sets_per_macro)

        self.set_addr_width = clog2(total_sets)

        # self.storage_addr_width = self.

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        # phases = [] TODO
        # Take in the valid and data and attach an address + direct to a port
        self._config_data_in = self.input("config_data_in",
                                          self.data_width)

        self._config_addr_in = self.input("config_addr_in",
                                          self.config_addr_width)

        self._config_wr = self.input("config_wr", 1)
        self._config_rd = self.input("config_rd", 1)
        self._config_en = self.input("config_en", self.total_sets)

        self._clk_en = self.input("clk_en", 1)

        self._rd_data_stg = self.input("rd_data_stg", self.data_width,
                                       size=(self.banks,
                                             self.fw_int),
                                       explicit_array=True,
                                       packed=True)

        self._wr_data = self.output("wr_data",
                                    self.data_width,
                                    size=self.fw_int,
                                    explicit_array=True,
                                    packed=True)

        self._rd_data_out = self.output("rd_data_out", self.data_width,
                                        size=self.total_sets,
                                        explicit_array=True,
                                        packed=True)

        self._addr_out = self.output("addr_out",
                                     self.addr_width)

        # One set per macro means we directly send the config address through
        if self.sets_per_macro == 1:
            width = self.addr_width - self.config_addr_width
            if width > 0:
                self.wire(self._addr_out, kts.concat(kts.const(0, width), self._config_addr_in))
            else:
                self.wire(self._addr_out, self._config_addr_in[self.addr_width - 1, 0])
        else:
            width = self.addr_width - self.config_addr_width - clog2(self.sets_per_macro)
            self._set_to_addr = self.var("set_to_addr",
                                         clog2(self.sets_per_macro))
            self._reduce_en = self.var("reduce_en", self.sets_per_macro)
            for i in range(self.sets_per_macro):
                reduce_var = self._config_en[i]
                for j in range(self.banks - 1):
                    reduce_var = kts.concat(reduce_var, self._config_en[i + (self.sets_per_macro * (j + 1))])
                self.wire(self._reduce_en[i], reduce_var.r_or())
            self.add_code(self.demux_set_addr)
            if width > 0:
                self.wire(self._addr_out, kts.concat(kts.const(0, width),
                          self._set_to_addr,
                          self._config_addr_in))
            else:
                self.wire(self._addr_out, kts.concat(self._set_to_addr, self._config_addr_in))

        self._wen_out = self.output("wen_out", self.banks)
        self._ren_out = self.output("ren_out", self.banks)

        # Handle data passing
        if self.fw_int == 1:
            # If word width is same as data width, just pass everything through
            self.wire(self._wr_data[0], self._config_data_in)
            # self.wire(self._rd_data_out, self._rd_data_stg[0])
            num = 0
            for i in range(self.banks):
                for j in range(self.sets_per_macro):
                    self.wire(self._rd_data_out[num], self._rd_data_stg[i])
                    num = num + 1
        else:
            self._data_wr_reg = self.var("data_wr_reg",
                                         self.data_width,
                                         size=self.fw_int - 1,
                                         packed=True,
                                         explicit_array=True)
            # self._data_rd_reg = self.var("data_rd_reg",
            #                              self.data_width,
            #                              size=self.fw_int - 1,
            #                              packed=True,
            #                              explicit_array=True)

            # Have word counter for repeated reads/writes
            self._cnt = self.var("cnt", clog2(self.fw_int))
            self._rd_cnt = self.var("rd_cnt", clog2(self.fw_int))
            self.add_code(self.update_cnt)
            self.add_code(self.update_rd_cnt)
            # Gate wen if not about to finish the word

            num = 0
            for i in range(self.banks):
                for j in range(self.sets_per_macro):
                    self.wire(self._rd_data_out[num], self._rd_data_stg[i][self._rd_cnt])
                    num = num + 1

            # Deal with writing to the data buffer
            self.add_code(self.write_buffer)

            # Wire the reg + such to this guy
            for i in range(self.fw_int - 1):
                self.wire(self._wr_data[i], self._data_wr_reg[i])
            self.wire(self._wr_data[self.fw_int - 1], self._config_data_in)

        # If we have one bank, we can just always rd/wr from that one
        if self.banks == 1:
            if self.fw_int == 1:
                self.wire(self._wen_out, self._config_wr)
            else:
                self.wire(self._wen_out,
                          self._config_wr & (self._cnt == (self.fw_int - 1)))
            self.wire(self._ren_out, self._config_rd)
        # Otherwise we need to extract the bank from the set
        else:
            if self.fw_int == 1:
                for i in range(self.banks):
                    width = self.sets_per_macro
                    self.wire(self._wen_out[i], self._config_wr &
                              self._config_en[(i + 1) * width - 1, i * width].r_or())
            else:
                for i in range(self.banks):
                    width = self.sets_per_macro
                    self.wire(self._wen_out[i],
                              self._config_wr & self._config_en[(i + 1) * width - 1, i * width].r_or() &
                              (self._cnt == (self.fw_int - 1)))

            for i in range(self.banks):
                width = self.sets_per_macro
                self.wire(self._ren_out[i],
                          self._config_rd & self._config_en[(i + 1) * width - 1, i * width].r_or())

    @always_comb
    def demux_set_addr(self):
        self._set_to_addr = 0
        for i in range(self.sets_per_macro):
            if self._reduce_en[i]:
                self._set_to_addr = i

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_cnt(self):
        if ~self._rst_n:
            self._cnt = 0
        # Increment when reading/writing - making sure
        # that the sequencing is correct from app level!
        elif (self._config_wr | self._config_rd) & self._config_en.r_or():
            self._cnt = self._cnt + 1

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_rd_cnt(self):
        if ~self._rst_n:
            self._rd_cnt = 0
        # Increment when reading/writing - making sure
        # that the sequencing is correct from app level!
        else:
            self._rd_cnt = self._cnt

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def write_buffer(self):
        if ~self._rst_n:
            self._data_wr_reg = 0
        # Increment when reading/writing - making sure
        # that the sequencing is correct from app level!
        elif self._config_wr & (self._cnt < self.fw_int - 1):
            self._data_wr_reg[self._cnt] = self._config_data_in


if __name__ == "__main__":
    db_dut = StorageConfigSeq(data_width=16,
                              config_addr_width=8,
                              addr_width=16,
                              fetch_width=64,
                              total_sets=2,
                              sets_per_macro=2)
    verilog(db_dut, filename="storage_config_seq.sv",
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.utils.util import safe_wire, add_counter, decode
import kratos as kts


class StrgUBTBOnly(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 tb_height=2):

        super().__init__("strg_ub_tb_only")

        ##################################################################################
        # Capture constructor parameter...
        ##################################################################################
        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.tb_height = tb_height
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.config_width = config_width
        self.data_width = data_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support

        self.default_iterator_support = 6
        self.default_config_width = 16
        self.sram_iterator_support = 6
        self.agg_rd_addr_gen_width = 8

        ##################################################################################
        # IO
        ##################################################################################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._cycle_count = self.input("cycle_count", 16)

        # data from SRAM
        self._sram_read_data = self.input("sram_read_data", self.data_width,
                                          size=self.fetch_width,
                                          packed=True,
                                          explicit_array=True)
        # read enable from SRAM
        self._t_read = self.input("t_read", self.interconnect_output_ports)

        # sram to tb for loop
        self._loops_sram2tb_mux_sel = self.input("loops_sram2tb_mux_sel",
                                                 width=max(clog2(self.default_iterator_support), 1),
                                                 size=self.interconnect_output_ports,
                                                 explicit_array=True,
                                                 packed=True)

        self._loops_sram2tb_restart = self.input("loops_sram2tb_restart",
                                                 width=1,
                                                 size=self.interconnect_output_ports,
                                                 explicit_array=True,
                                                 packed=True)

        self._valid_out = self.output("accessor_output", self.interconnect_output_ports)
        self._data_out = self.output("data_out", self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        ##################################################################################
        # TB RELEVANT SIGNALS
        ##################################################################################
        self._tb = self.var("tb",
                            width=self.data_width,
                            size=(self.interconnect_output_ports,
                                  self.tb_height,
                                  self.fetch_width),
                            packed=True,
                            explicit_array=True)

        self._tb_write_addr = self.var("tb_write_addr", 2 + max(1, clog2(self.tb_height)),
                                       size=self.interconnect_output_ports,
                                       packed=True,
                                       explicit_array=True)

        self._tb_read_addr = self.var("tb_read_addr", 2 + max(1, clog2(self.tb_height)),
                                      size=self.interconnect_output_ports,
                                      packed=True,
                                      explicit_array=True)

        # write enable to tb, delayed 1 cycle from SRAM reads
        self._t_read_d1 = self.var("t_read_d1", self.interconnect_output_ports)
        # read enable for reads from tb
        self._tb_read = self.var("tb_read", self.interconnect_output_ports)

        # Break out valids...
        self.wire(self._valid_out, self._tb_read)

        # delayed input mux_sel and restart signals from sram read/tb write
        # for loop and scheduling
        self._mux_sel_d1 = self.var("mux_sel_d1",
                                    kts.clog2(self.default_iterator_support),
                                    size=self.interconnect_output_ports,
                                    packed=True,
                                    explicit_array=True)

        self._restart_d1 = self.var("restart_d1",
                                    width=1,
                                    size=self.interconnect_output_ports,
                                    explicit_array=True,
                                    packed=True)

        for i in range(self.interconnect_output_ports):
            # signals delayed by 1 cycle from SRAM
            @always_ff((posedge, "clk"), (negedge, "rst_n"))
            def delay_read():
                if ~self._rst_n:
                    self._t_read_d1[i] = 0
                    self._mux_sel_d1[i] = 0
                    self._restart_d1[i] = 0
                else:
                    self._t_read_d1[i] = self._t_read[i]
                    self._mux_sel_d1[i] = self._loops_sram2tb_mux_sel[i]
                    self._restart_d1[i] = self._loops_sram2tb_restart[i]
            self.add_code(delay_read)

        ##################################################################################
        # TB PATHS
        ##################################################################################
        for i in range(self.interconnect_output_ports):

            self.tb_iter_support = 6
            self.tb_addr_width = 4
            self.tb_range_width = 16

            _AG = AddrGen(iterator_support=self.default_iterator_support,
                          config_width=self.tb_addr_width)

            self.add_child(f"tb_write_addr_gen_{i}",
                           _AG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._t_read_d1[i],
                           mux_sel=self._mux_sel_d1[i],
                           restart=self._restart_d1[i])
            safe_wire(gen=self, w_to=self._tb_write_addr[i], w_from=_AG.ports.addr_out)

            @always_ff((posedge, "clk"))
            def tb_ctrl():
                if self._t_read_d1[i]:
                    self._tb[i][self._tb_write_addr[i][0]] = \
                        self._sram_read_data
            self.add_code(tb_ctrl)

            # READ FROM TB

            fl_ctr_tb_rd = ForLoop(iterator_support=self.tb_iter_support,
                                   config_width=self.tb_range_width)

            self.add_child(f"loops_buf2out_read_{i}",
                           fl_ctr_tb_rd,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._tb_read[i])

            _AG = AddrGen(iterator_support=self.tb_iter_support,
                          config_width=self.tb_addr_width)
            self.add_child(f"tb_read_addr_gen_{i}",
                           _AG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._tb_read[i],
                           # addr_out=self._tb_read_addr[i])
                           mux_sel=fl_ctr_tb_rd.ports.mux_sel_out,
                           restart=fl_ctr_tb_rd.ports.restart)
            safe_wire(gen=self, w_to=self._tb_read_addr[i], w_from=_AG.ports.addr_out)

            self.add_child(f"tb_read_sched_gen_{i}",
                           SchedGen(iterator_support=self.tb_iter_support,
                                    # config_width=self.tb_addr_width),
                                    config_width=16),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           cycle_count=self._cycle_count,
                           mux_sel=fl_ctr_tb_rd.ports.mux_sel_out,
                           finished=fl_ctr_tb_rd.ports.restart,
                           valid_output=self._tb_read[i])

            @always_comb
            def tb_to_out():
                self._data_out[i] = self._tb[i][self._tb_read_addr[i][clog2(self.tb_height) +
                                                                      clog2(self.fetch_width) - 1,
                                                                      clog2(self.fetch_width)]][self._tb_read_addr[i][clog2(self.fetch_width) - 1, 0]]
            self.add_code(tb_to_out)


if __name__ == "__main__":
    lake_dut = StrgUBTBOnly()
    verilog(lake_dut, filename="strg_ub_tb_only.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.aggregator import Aggregator
from lake.modules.addr_gen import AddrGen
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg


class OutputAddrCtrl(Generator):
    '''
    Input addressing control from the aggregation buffers to the SRAM
    '''
    def __init__(self,
                 interconnect_output_ports,
                 mem_depth,
                 num_tiles,
                 banks,
                 iterator_support,
                 address_width,
                 config_width=16):
        super().__init__("output_addr_ctrl")

        self.interconnect_output_ports = interconnect_output_ports
        self.mem_depth = mem_depth
        self.num_tiles = num_tiles
        self.banks = banks
        self.iterator_support = iterator_support
        self.address_width = address_width
        self.port_sched_width = clog2(self.interconnect_output_ports)
        self.config_width = config_width

        self.mem_addr_width = clog2(self.num_tiles * self.mem_depth)

        if self.banks > 1:
            self.bank_addr_width = clog2(self.banks)
        else:
            self.bank_addr_width = 0
        self.address_width = self.mem_addr_width + self.bank_addr_width

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        # Take in the valid and attach an address + direct to a port
        self._valid_in = self.input("valid_in", self.interconnect_output_ports)
        self._step_in = self.input("step_in", self.interconnect_output_ports)

        # Outputs
        self._ren = self.output("ren",
                                self.interconnect_output_ports,
                                size=self.banks,
                                explicit_array=True,
                                packed=True)

        self._addresses = self.output("addr_out",
                                      self.mem_addr_width,
                                      size=self.interconnect_output_ports,
                                      explicit_array=True,
                                      packed=True)

        # LOCAL VARS
        self._local_addrs = self.var("local_addrs",
                                     self.address_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        if self.banks == 1 and self.interconnect_output_ports == 1:
            self.wire(self._ren[0][0], self._valid_in)
        elif self.banks == 1 and self.interconnect_output_ports > 1:
            self.add_code(self.set_ren_single)
        elif self.banks > 1 and self.interconnect_output_ports == 1:
            self.add_code(self.set_ren_mult)
        else:
            self.add_code(self.set_ren_mult)

        # MAIN

        self.add_code(self.set_out)

        # Now we should instantiate the child address generators
        # (1 per input port) to send to the sram banks
        for i in range(self.interconnect_output_ports):
            new_addr_gen = AddrGen(iterator_support=self.iterator_support,
                                   # address_width=self.address_width,
                                   config_width=self.config_width)

            self.add_child(f"address_gen_{i}", new_addr_gen,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           clk_en=const(1, 1),
                           flush=const(0, 1))
            # self.add_stmt(new_addr_gen.ports.step.assign(self._valid_in[i]))
            self.add_stmt(new_addr_gen.ports.step.assign(self._step_in[i] & self._valid_in[i]))

            # Get the address for each input port
            self.wire(self._local_addrs[i],
                      new_addr_gen.ports.addr_out[self.address_width - 1, 0])

    @always_comb
    def set_ren_mult(self):
        self._ren = 0
        for i in range(self.interconnect_output_ports):
            if(self._valid_in[i]):
                self._ren[self._local_addrs[i][self.mem_addr_width + self.bank_addr_width - 1,
                                               self.mem_addr_width]][i] = 1

    @always_comb
    def set_ren_single(self):
        self._ren = 0
        for i in range(self.interconnect_output_ports):
            if(self._valid_in[i]):
                self._ren[0][i] = 1

    @always_comb
    def set_out(self):
        self._addresses = 0
        for i in range(self.interconnect_output_ports):
            self._addresses[i] = self._local_addrs[i][self.mem_addr_width - 1, 0]


if __name__ == "__main__":
    # lifting ports up

    db_dut = OutputAddrCtrl(interconnect_output_ports=2,
                            mem_depth=512,
                            num_tiles=2,
                            banks=2,
                            iterator_support=6,
                            address_width=16)
    verilog(db_dut, filename="output_addr_ctrl.sv",
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.sram_stub import SRAMStub
from lake.modules.two_port_sram_stub import TwoPortSRAMStub
from lake.modules.pipe_reg import PipeReg
from lake.modules.virtual_remap_table import VirtualRemapTable
from math import *


class SinglePortMemory(Generator):
    def __init__(self,
                 macro_depth,
                 macro_width,
                 desired_depth,
                 desired_width):

        super().__init__("single_port_memory")

        # PORT DEFS: begin
        self._clk = self.clock("i_clk")
        self._rst_n = self.reset("i_rst_n")

        self.full_addr = clog2(desired_depth)
        self.addr_width_macro = clog2(macro_depth)

        self._addr = self.input("i_addr", self.full_addr)
        self._rd_data = self.output("o_rd_data", desired_width)
        self._ren = self.input("i_ren", 1)

        self._wr_data = self.input("i_wr_data", desired_width)
        self._wen = self.input("i_wen", 1)
        # PORT DEFS: end

        banks_wide = int(desired_width / macro_width)
        banks_tall = int(desired_depth / macro_depth)
        print(f"Generating memory with {banks_wide} banks for width")
        print(f"and {banks_tall} banks for capacity")

        # LOCAL SIGNALS: begin
        # This is the output of all banks (tallwise)
        self._memory_space = self.var("memory_space",
                                      width=desired_width,
                                      size=banks_tall,
                                      packed=True)
        # Need to reg the rd addr to know how to select the data on the next cycle
        self._bank_sel = self.var("bank_sel",
                                  max(self.full_addr - self.addr_width_macro, 1))
        self._bank_sel_reg = self.var("bank_sel_reg",
                                      max(self.full_addr - self.addr_width_macro, 1))
        # LOCAL_SIGNALS: end

        # GENERATION LOGIC: begin

        # If already a two-port, there is only one cycle of latency

        # Either the difference in address widths or 1 bit if there's no difference

        if(banks_tall > 1):
            self.wire(self._bank_sel, self._addr[self.full_addr - 1, self.addr_width_macro])
        else:
            self.wire(self._bank_sel, 0)
        # Update the bank sel reg
        self.add_code(self.always_bank_sel_reg)

        self.instantiate_single_port(banks_tall, banks_wide, macro_width, macro_depth)

        # GENERATION LOGIC: end
        # still want passthru_block for read+write to same location ?

    def instantiate_single_port(self,
                                banks_tall,
                                banks_wide,
                                macro_width,
                                macro_depth):
        '''
        Hook up the set of single port memories
        '''

        self._sub_addr = self.var("addr_macro", self.addr_width_macro)
        self.wire(self._sub_addr, self._addr[self.addr_width_macro - 1, 0])

        # Hook up the grid of SRAM banks
        for i in range(banks_tall):
            for j in range(banks_wide):
                self.add_child(f"sram_d{i}_w{j}", SRAMStub(macro_width, macro_depth))

                self.wire(self[f"sram_d{i}_w{j}"].ports.i_clk, self._clk)
                self.wire(self[f"sram_d{i}_w{j}"].ports.i_rst_n, self._rst_n)

                self.wire(self[f"sram_d{i}_w{j}"].ports.i_addr, self._sub_addr)
                if(banks_tall > 1):
                    self.wire(self[f"sram_d{i}_w{j}"].ports.i_wen,
                              self._wen & (self._addr[self.full_addr - 1,
                                                      self.addr_width_macro] == i))
                    self.wire(self[f"sram_d{i}_w{j}"].ports.i_cen,
                              (self._ren | self._wen) & (self._addr[self.full_addr - 1,
                                                         self.addr_width_macro] == i))
                else:
                    self.wire(self[f"sram_d{i}_w{j}"].ports.i_wen, self._wen)
                    self.wire(self[f"sram_d{i}_w{j}"].ports.i_cen, (self._ren | self._wen))

                # Send the data across horizontally based on j and the macro width
                self.wire(self[f"sram_d{i}_w{j}"].ports.i_data,
                          self._wr_data[((j + 1) * macro_width) - 1, j * macro_width])

        for i in range(banks_tall):
            if(banks_wide > 1):
                self.wire(self._memory_space[i],
                          concat(*[self[f"sram_d{i}_w{j}"].ports.o_data
                                   for j in range(banks_wide)]))
            else:
                self.wire(self._memory_space, self[f"sram_d{i}_w{j}"].ports.o_data)

        if(banks_tall > 1):
            self.wire(self._rd_data, self._memory_space[self._bank_sel_reg])
        else:
            self.wire(self._rd_data, self._memory_space)

        return 0

    @always((posedge, "i_clk"), (negedge, "i_rst_n"))
    def always_bank_sel_reg(self):
        if ~self._rst_n:
            self._bank_sel_reg = 0
        elif self._ren:
            self._bank_sel_reg = self._bank_sel


if __name__ == "__main__":
    tpm_dut = SinglePortMemory(macro_depth=128, macro_width=16, desired_depth=256, desired_width=32)
    verilog(tpm_dut, filename="single_port_memory.sv", check_active_high=False)



from kratos import *
from math import log


class RegisterFile(Generator):

    ##########################
    # Generation             #
    ##########################
    def __init__(self,
                 data_width,
                 write_ports,
                 read_ports,
                 width_mult,
                 depth,
                 read_delay=0
                 ):
        super().__init__("register_file")

        self.width_mult = width_mult
        self.data_width = data_width
        self.write_ports = write_ports
        self.read_ports = read_ports
        self.depth = depth
        self.read_delay = read_delay

        ############################
        # Clock and Reset          #
        ############################
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        ############################
        # Inputs                   #
        ############################

        if self.read_ports == 1:
            self._rd_addr = self.input("rd_addr", clog2(self.depth))
            self._data_out = self.output("data_out", self.data_width,
                                         size=self.width_mult,
                                         explicit_array=True,
                                         packed=True)
        else:
            self._rd_addr = self.input("rd_addr", clog2(self.depth),
                                       size=self.read_ports,
                                       explicit_array=True,
                                       packed=True)
            self._data_out = self.output("data_out",
                                         self.data_width,
                                         size=(self.read_ports,
                                               self.width_mult),
                                         explicit_array=True,
                                         packed=True)

        self._wen = self.input("wen", self.write_ports)
        if self.write_ports == 1:
            self._wr_addr = self.input("wr_addr", clog2(self.depth))
            self._data_in = self.input("data_in",
                                       self.data_width,
                                       size=self.width_mult,
                                       explicit_array=True,
                                       packed=True)
        else:
            self._wr_addr = self.input("wr_addr", clog2(self.depth),
                                       size=self.write_ports,
                                       explicit_array=True,
                                       packed=True)
            self._data_in = self.input("data_in",
                                       self.data_width,
                                       size=(self.write_ports,
                                             self.width_mult),
                                       explicit_array=True,
                                       packed=True)

        ############################
        # Local Variables          #
        ############################
        self._data_array = self.var("data_array",
                                    self.data_width,
                                    size=(self.depth,
                                          self.width_mult),
                                    packed=True,
                                    explicit_array=True)

        ############################
        # Add seq blocks           #
        ############################
        if self.write_ports == 1:
            self.add_code(self.seq_data_access_one_w)
        else:
            self.add_code(self.seq_data_access)

        if self.read_ports == 1:
            if self.read_delay == 1:
                self._ren = self.input("ren", 1)
                self.add_code(self.seq_data_out_one_r)
            else:
                self.add_code(self.comb_data_out_one_r)
        else:
            if self.read_delay == 1:
                self.add_code(self.seq_data_out)
            else:
                self.add_code(self.comb_data_out)

    ##########################
    # Access sram array      #
    ##########################
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def seq_data_access(self):
        for i in range(self.write_ports):
            if ~self._rst_n:
                self._data_array = 0
            elif self._wen[i]:
                self._data_array[self._wr_addr[i]] = self._data_in[i]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def seq_data_access_one_w(self):
        if ~self._rst_n:
            self._data_array = 0
        elif self._wen:
            self._data_array[self._wr_addr] = self._data_in

    @always_comb
    def comb_data_out(self):
        for i in range(self.read_ports):
            self._data_out[i] = self._data_array[self._rd_addr[i]]

    @always_comb
    def comb_data_out_one_r(self):
        self._data_out = self._data_array[self._rd_addr]

    @always_ff((posedge, "clk"))
    def seq_data_out(self):
        for i in range(self.read_ports):
            if self._ren:
                self._data_out[i] = self._data_array[self._rd_addr[i]]

    @always_ff((posedge, "clk"))
    def seq_data_out_one_r(self):
        if self._ren:
            self._data_out = self._data_array[self._rd_addr]


if __name__ == "__main__":
    dut = RegisterFile(16, 1, 2, 1, 64)
    verilog(dut, filename="register_file.sv", check_flip_flop_always_ff=False)



from lake.models.model import Model


class RegisterFileModel(Model):
    def __init__(self,
                 data_width,
                 write_ports,
                 read_ports,
                 width_mult,
                 depth):

        self.width_mult = width_mult
        self.data_width = data_width
        self.write_ports = write_ports
        self.read_ports = read_ports
        self.depth = depth

        self.mem = []
        for i in range(self.depth):
            row = []
            for j in range(self.width_mult):
                row.append(0)
            self.mem.append(row)

    def set_config(self, new_config):
        # No configuration space
        return

    def interact(self,
                 wen,
                 wr_addr,
                 rd_addr,
                 data_in):
        '''
        Returns rd_data (list)
        '''

        # Do the read first - data won't pass through on the same cycle

        if type(rd_addr) is int:
            rd_addr = [rd_addr]
        if type(wr_addr) is int:
            wr_addr = [wr_addr]
        if type(wen) is int:
            wen = [wen]
        if type(data_in) is int:
            data_in = [data_in]
        elif type(data_in[0]) is int:
            data_in = [data_in]

        ret_data = []
        for i in range(self.read_ports):
            ret_data.append(self.mem[rd_addr[i]].copy())

        for i in range(self.write_ports):
            if wen[i] == 1:
                self.mem[wr_addr[i]] = data_in[i].copy()

        return ret_data.copy()

    def get_reads(self, rd_addr):
        ret_data = []
        for i in range(self.read_ports):
            ret_data.append(self.mem[rd_addr[i]].copy())

    def dump_mem(self):
        for i in range(self.depth):
            print(f"addr: {i}, data: {self.mem[i]}")



import kratos as kts
from enum import Enum


class RangeGroupAttr(kts.Attribute):
    def __init__(self,
                 group_num,
                 group_name,
                 doc_string=""):
        super().__init__()
        self.value = "range_group"
        self.group_num = group_num
        self.group_name = group_name
        self.documentation = doc_string

    def set_documentation(self, new_doc):
        self.documentation = new_doc

    def get_documentation(self):
        return self.documentation

    def get_group_num(self):
        return self.group_num

    def geT_group_name(self):
        return self.group_name



from lake.models.model import Model
from lake.models.agg_aligner_model import AggAlignerModel
from lake.models.agg_buff_model import AggBuffModel
from lake.models.input_addr_ctrl_model import InputAddrCtrlModel
from lake.models.output_addr_ctrl_model import OutputAddrCtrlModel
from lake.models.rw_arbiter_model import RWArbiterModel
from lake.models.demux_reads_model import DemuxReadsModel
from lake.models.sync_groups_model import SyncGroupsModel
from lake.models.prefetcher_model import PrefetcherModel
from lake.models.tba_model import TBAModel
from lake.models.register_file_model import RegisterFileModel
from lake.models.app_ctrl_model import AppCtrlModel
from lake.models.sram_wrapper_model import SRAMWrapperModel
# from lake.models.chain_model import ChainModel
from typing import List
import math as mt
import kratos as kts


class LakeTopModel(Model):

    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=2,
                 input_iterator_support=6,  # Addr Controllers
                 output_iterator_support=6,
                 interconnect_input_ports=1,  # Connection to int
                 interconnect_output_ports=3,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 use_sram_stub=1,
                 sram_name="default_name",
                 read_delay=1,
                 agg_height=8,
                 max_agg_schedule=64,
                 input_max_port_sched=64,
                 output_max_port_sched=64,
                 align_input=1,
                 max_line_length=2048,
                 tb_height=1,
                 tb_range_max=2048,
                 tb_range_inner_max=5,
                 tb_sched_max=64,
                 num_tb=1,
                 multiwrite=2,
                 max_prefetch=64,
                 num_tiles=1,
                 stcl_valid_iter=4):

        self.data_width = data_width
        self.mem_width = mem_width
        self.mem_depth = mem_depth
        self.banks = banks
        self.input_iterator_support = input_iterator_support
        self.output_iterator_support = output_iterator_support
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.mem_input_ports = mem_input_ports
        self.mem_output_ports = mem_output_ports
        self.use_sram_stub = use_sram_stub
        self.sram_name = sram_name
        self.agg_height = agg_height
        self.max_agg_schedule = max_agg_schedule
        self.input_max_port_sched = input_max_port_sched
        self.output_max_port_sched = output_max_port_sched
        self.input_port_sched_width = kts.clog2(self.interconnect_input_ports)
        self.align_input = align_input
        self.max_line_length = max_line_length
        assert self.mem_width > self.data_width, "Data width needs to be smaller than mem"
        self.fw_int = int(self.mem_width / self.data_width)
        self.num_tb = num_tb
        self.tb_height = tb_height
        self.tb_range_max = tb_range_max
        self.tb_range_inner_max = tb_range_inner_max
        self.tb_sched_max = tb_sched_max
        self.multiwrite = multiwrite
        self.max_prefetch = max_prefetch
        self.read_delay = read_delay
        self.num_tiles = num_tiles
        self.stcl_valid_iter = stcl_valid_iter

        self.chain_idx_bits = max(1, kts.clog2(num_tiles))
        self.address_width = kts.clog2(self.mem_depth * num_tiles)

        self.config = {}

        # top level configuration registers

        # chaining
        self.config[f"enable_chain_input"] = 0
        self.config[f"enable_chain_output"] = 0
        self.config[f"chain_idx_input"] = 0
        self.config[f"chain_idx_output"] = 0

        self.config[f"tile_en"] = 0
        self.config[f"mode"] = 0
        for i in range(self.interconnect_output_ports):
            self.config[f"rate_matched_{i}"] = 0

        # Set up model..
        ### APP CTRL (FINE-GRAINED)
        self.app_ctrl = AppCtrlModel(int_in_ports=self.interconnect_input_ports,
                                     int_out_ports=self.interconnect_output_ports,
                                     sprt_stcl_valid=True,
                                     stcl_iter_support=self.stcl_valid_iter)
        for i in range(self.interconnect_input_ports):
            self.config[f"app_ctrl_write_depth_{i}"] = 0
        for i in range(self.interconnect_output_ports):
            self.config[f"app_ctrl_input_port_{i}"] = 0
            self.config[f"app_ctrl_read_depth_{i}"] = 0
            self.config[f"app_ctrl_prefill_{i}"] = 0

        # calculate stencil_valid with fine-grained application controller
        for i in range(self.stcl_valid_iter):
            self.config[f'app_ctrl_ranges_{i}'] = 0
            self.config[f'app_ctrl_app_ctrl_threshold_{i}'] = 0

        ### COARSE APP CTRL
        self.app_ctrl_coarse = AppCtrlModel(int_in_ports=self.interconnect_input_ports,
                                            int_out_ports=self.interconnect_output_ports,
                                            sprt_stcl_valid=False,
                                            # unused, stencil valid comes from app ctrl,
                                            # not coarse app ctrl
                                            stcl_iter_support=0)

        for i in range(self.interconnect_input_ports):
            self.config[f"app_ctrl_coarse_write_depth_{i}"] = 0
        for i in range(self.interconnect_output_ports):
            self.config[f"app_ctrl_coarse_input_port_{i}"] = 0
            self.config[f"app_ctrl_coarse_read_depth_{i}"] = 0
            self.config[f"app_ctrl_coarse_prefill_{i}"] = 0

        ### INST AGG ALIGNER
        if(self.agg_height > 0):
            self.agg_aligners = []
            for i in range(self.interconnect_input_ports):
                self.agg_aligners.append(AggAlignerModel(data_width=self.data_width,
                                                         max_line_length=self.max_line_length))
                self.config[f"agg_align_{i}_line_length"] = 0

        ### AGG BUFF
        self.agg_buffs = []
        for port in range(self.interconnect_input_ports):
            self.agg_buffs.append(AggBuffModel(agg_height=self.agg_height,
                                               data_width=self.data_width,
                                               mem_width=self.mem_width,
                                               max_agg_schedule=self.max_agg_schedule))

            self.config[f"agg_in_{i}_in_period"] = 0
            self.config[f"agg_in_{i}_out_period"] = 0
            for j in range(self.max_agg_schedule):
                self.config[f"agg_in_{i}_in_sched_{j}"] = 0
                self.config[f"agg_in_{i}_out_sched_{j}"] = 0

        ### INPUT ADDR CTRL
        self.iac = InputAddrCtrlModel(interconnect_input_ports=self.interconnect_input_ports,
                                      data_width=self.data_width,
                                      fetch_width=self.mem_width,
                                      mem_depth=self.mem_depth,
                                      num_tiles=self.num_tiles,
                                      banks=self.banks,
                                      iterator_support=self.input_iterator_support,
                                      max_port_schedule=self.input_max_port_sched,
                                      address_width=self.address_width)

        for i in range(self.interconnect_input_ports):
            self.config[f"input_addr_ctrl_address_gen_{i}_dimensionality"] = 0
            self.config[f"input_addr_ctrl_address_gen_{i}_starting_addr"] = 0
            for j in range(self.input_iterator_support):
                self.config[f"input_addr_ctrl_address_gen_{i}_ranges_{j}"] = 0
                self.config[f"input_addr_ctrl_address_gen_{i}_strides_{j}"] = 0
            for j in range(self.multiwrite):
                self.config[f"input_addr_ctrl_offsets_cfg_{i}_{j}"] = 0

        ### OUTPUT ADDR CTRL
        self.oac = OutputAddrCtrlModel(interconnect_output_ports=self.interconnect_output_ports,
                                       mem_depth=self.mem_depth,
                                       num_tiles=self.num_tiles,
                                       data_width=self.data_width,
                                       fetch_width=self.mem_width,
                                       banks=self.banks,
                                       iterator_support=self.output_iterator_support,
                                       address_width=self.address_width,
                                       chain_idx_output=self.config[f"chain_idx_output"])
        for i in range(self.interconnect_output_ports):
            self.config[f"output_addr_ctrl_address_gen_{i}_dimensionality"] = 0
            self.config[f"output_addr_ctrl_address_gen_{i}_starting_addr"] = 0
            for j in range(self.input_iterator_support):
                self.config[f"output_addr_ctrl_address_gen_{i}_ranges_{j}"] = 0
                self.config[f"output_addr_ctrl_address_gen_{i}_strides_{j}"] = 0

        ### RW ARBITER
        # Per bank allocation
        self.rw_arbs = []
        for bank in range(self.banks):
            self.rw_arbs.append(RWArbiterModel(fetch_width=self.mem_width,
                                               data_width=self.data_width,
                                               memory_depth=self.mem_depth,
                                               int_out_ports=self.interconnect_output_ports,
                                               read_delay=self.read_delay))

        self.mems = []
        if self.read_delay == 1:
            ### SRAMS
            for banks in range(self.banks):
                self.mems.append(SRAMWrapperModel(use_sram_stub=self.use_sram_stub,
                                                  sram_name=self.sram_name,
                                                  data_width=self.data_width,
                                                  fw_int=self.fw_int,
                                                  mem_depth=self.mem_depth,
                                                  mem_input_ports=self.mem_input_ports,
                                                  mem_output_ports=self.mem_output_ports,
                                                  address_width=self.address_width,
                                                  bank_num=banks,
                                                  num_tiles=self.num_tiles,
                                                  enable_chain_input=self.config[f"enable_chain_input"],
                                                  enable_chain_output=self.config[f"enable_chain_output"],
                                                  chain_idx_input=self.config[f"chain_idx_input"],
                                                  chain_idx_output=self.config[f"chain_idx_output"]))
        else:
            ### REGFILES
            for banks in range(self.banks):
                self.mems.append(RegisterFileModel(data_width=self.data_width,
                                                   write_ports=self.mem_input_ports,
                                                   read_ports=self.mem_output_ports,
                                                   width_mult=self.fw_int,
                                                   depth=self.mem_depth))

        ### DEMUX READS
        self.demux_reads = DemuxReadsModel(fetch_width=self.mem_width,
                                           data_width=self.data_width,
                                           banks=self.banks,
                                           int_out_ports=self.interconnect_output_ports)

        ### SYNC GROUPS
        self.sync_groups = SyncGroupsModel(fetch_width=self.mem_width,
                                           data_width=self.data_width,
                                           int_out_ports=self.interconnect_output_ports)
        for i in range(self.interconnect_output_ports):
            self.config[f"sync_grp_sync_group_{i}"] = 0

        ### PREFETCHERS
        self.prefetchers = []
        for port in range(self.interconnect_output_ports):
            self.prefetchers.append(PrefetcherModel(fetch_width=self.mem_width,
                                                    data_width=self.data_width,
                                                    max_prefetch=self.max_prefetch))
            self.config[f"pre_fetch_{port}_input_latency"] = 0

        ### TBAS
        self.tbas = []
        for port in range(self.interconnect_output_ports):
            self.tbas.append(TBAModel(word_width=self.data_width,
                                      fetch_width=self.fw_int,
                                      num_tb=self.num_tb,
                                      tb_height=self.tb_height,
                                      max_range=self.tb_range_max,
                                      max_range_inner=self.tb_range_inner_max))
            for i in range(self.tb_height):
                self.config[f"tba_{port}_tb_{i}_range_inner"] = 0
                self.config[f"tba_{port}_tb_{i}_range_outer"] = 0
                self.config[f"tba_{port}_tb_{i}_stride"] = 0
                self.config[f"tba_{port}_tb_{i}_dimensionality"] = 0
                self.config[f"tba_{port}_tb_{i}_starting_addr"] = 0
                for j in range(self.tb_sched_max):
                    self.config[f"tba_{port}_tb_{i}_indices_{j}"] = 0

        ### CHAINING
        # self.chain = ChainModel(data_width=self.data_width,
        #                         interconnect_output_ports=self.interconnect_output_ports,
        #                         chain_idx_bits=self.chain_idx_bits,
        #                         enable_chain_output=self.config[f"enable_chain_output"],
        #                         chain_idx_output=self.config[f"chain_idx_output"])
        # chaining configuration registers are passed down from top level

    def set_config(self, new_config):
        # Configure top level
        for key, config_val in new_config.items():
            edited_key = key
            # Hack for now to not need to reconfigure everything
            if "strg_ub_" in key:
                edited_key = key[8:]
            assert edited_key in self.config, f"{edited_key} is not in the configuration..."
            self.config[edited_key] = config_val

        # Configure children

        # Config App Ctrl
        app_ctrl_cfg = {}
        for i in range(self.interconnect_input_ports):
            app_ctrl_cfg[f"write_depth_{i}"] = self.config[f"app_ctrl_write_depth_{i}"]
        for i in range(self.interconnect_output_ports):
            app_ctrl_cfg[f"read_depth_{i}"] = self.config[f"app_ctrl_read_depth_{i}"]
            app_ctrl_cfg[f"input_port_{i}"] = self.config[f"app_ctrl_input_port_{i}"]
            app_ctrl_cfg[f"prefill_{i}"] = self.config[f"app_ctrl_prefill_{i}"]

        for i in range(self.stcl_valid_iter):
            app_ctrl_cfg[f"ranges_{i}"] = self.config[f'app_ctrl_ranges_{i}']
            app_ctrl_cfg[f"threshold_{i}"] = self.config[f'app_ctrl_app_ctrl_threshold_{i}']
        self.app_ctrl.set_config(app_ctrl_cfg)

        # Config Coarse App Ctrl
        app_ctrl_coarse_cfg = {}
        for i in range(self.interconnect_input_ports):
            app_ctrl_coarse_cfg[f"write_depth_{i}"] = self.config[f"app_ctrl_coarse_write_depth_{i}"]
        for i in range(self.interconnect_output_ports):
            app_ctrl_coarse_cfg[f"read_depth_{i}"] = self.config[f"app_ctrl_coarse_read_depth_{i}"]
            app_ctrl_coarse_cfg[f"input_port_{i}"] = self.config[f"app_ctrl_coarse_input_port_{i}"]
            app_ctrl_coarse_cfg[f"prefill_{i}"] = self.config[f"app_ctrl_coarse_prefill_{i}"]
        self.app_ctrl_coarse.set_config(app_ctrl_coarse_cfg)

        # Config Agg Align
        for i in range(self.interconnect_input_ports):
            agg_aligner_config = {}
            agg_aligner_config['line_length'] = self.config[f"agg_align_{i}_line_length"]
            self.agg_aligners[i].set_config(agg_aligner_config)

        # Config Agg Buff
        for i in range(self.interconnect_input_ports):
            agg_buff_config = {}
            agg_buff_config['in_period'] = self.config[f"agg_in_{i}_in_period"]
            agg_buff_config['out_period'] = self.config[f"agg_in_{i}_out_period"]
            for j in range(self.max_agg_schedule):
                agg_buff_config[f"in_sched_{j}"] = self.config[f"agg_in_{i}_in_sched_{j}"]
                agg_buff_config[f"out_sched_{j}"] = self.config[f"agg_in_{i}_out_sched_{j}"]
            self.agg_buffs[i].set_config(agg_buff_config)

        # Config IAC
        iac_config = {}
        for i in range(self.interconnect_input_ports):
            iac_config[f"address_gen_{i}_starting_addr"] = \
                self.config[f"input_addr_ctrl_address_gen_{i}_starting_addr"]
            iac_config[f"address_gen_{i}_dimensionality"] = \
                self.config[f"input_addr_ctrl_address_gen_{i}_dimensionality"]
            for j in range(self.input_iterator_support):
                iac_config[f"address_gen_{i}_ranges_{j}"] = \
                    self.config[f"input_addr_ctrl_address_gen_{i}_ranges_{j}"]
                iac_config[f"address_gen_{i}_strides_{j}"] = \
                    self.config[f"input_addr_ctrl_address_gen_{i}_strides_{j}"]
        self.iac.set_config(iac_config)

        # Config OAC
        oac_config = {}
        for i in range(self.interconnect_output_ports):
            oac_config[f"address_gen_{i}_starting_addr"] = \
                self.config[f"output_addr_ctrl_address_gen_{i}_starting_addr"]
            oac_config[f"address_gen_{i}_dimensionality"] = \
                self.config[f"output_addr_ctrl_address_gen_{i}_dimensionality"]
            for j in range(self.output_iterator_support):
                oac_config[f"address_gen_{i}_ranges_{j}"] = \
                    self.config[f"output_addr_ctrl_address_gen_{i}_ranges_{j}"]
                oac_config[f"address_gen_{i}_strides_{j}"] = \
                    self.config[f"output_addr_ctrl_address_gen_{i}_strides_{j}"]
        self.oac.set_config(oac_config)

        # Config RWArbiter
        rw_arb_config = {}
        for i in range(self.banks):
            self.rw_arbs[i].set_config(rw_arb_config)

        # Config SRAM
        sram_config = {}
        for i in range(self.banks):
            self.mems[i].set_config(sram_config)

        # Config DEMUX Reads
        demux_reads_config = {}
        self.demux_reads.set_config(demux_reads_config)

        # Config Sync Groups
        sync_groups_config = {}
        for i in range(self.interconnect_output_ports):
            sync_groups_config[f'sync_group_{i}'] = self.config[f"sync_grp_sync_group_{i}"]
        self.sync_groups.set_config(sync_groups_config)

        # Config Prefetcher
        for i in range(self.interconnect_output_ports):
            prefetch_config = {}
            prefetch_config['input_latency'] = self.config[f"pre_fetch_{i}_input_latency"]
            self.prefetchers[i].set_config(prefetch_config)

        # Config TBA
        for port in range(self.interconnect_output_ports):
            tba_config = {}
            # for i in range(self.tb_height):
            tba_config[f"range_inner"] = self.config[f"tba_{port}_tb_0_range_inner"]
            tba_config[f"range_outer"] = self.config[f"tba_{port}_tb_0_range_outer"]
            tba_config[f"stride"] = self.config[f"tba_{port}_tb_0_stride"]
            tba_config["indices"] = []
            tba_config[f"dimensionality"] = self.config[f"tba_{port}_tb_0_dimensionality"]
            tba_config[f"starting_addr"] = self.config[f"tba_{port}_tb_0_starting_addr"]
            for j in range(self.tb_sched_max):
                tba_config[f"indices"].append(self.config[f"tba_{port}_tb_0_indices_{j}"])
            self.tbas[port].set_config(tba_config)

        self.tb_valid = [0] * self.interconnect_output_ports

        self.mem_data_out = []
        self.mem_valid_data = []
        for i in range(self.banks):
            self.mem_data_out.append(0)
            self.mem_valid_data.append(0)

        self.iac_port_out = []
        self.valid_agg_buff = []
        for i in range(self.interconnect_input_ports):
            self.iac_port_out.append(0)
            self.valid_agg_buff.append(0)

    def interact(self,
                 chain_valid_in,
                 chain_data_in,
                 data_in,
                 addr_in,
                 wen,
                 ren):
        '''
        Top level interactions - - -
        returns (chain_data_out, chain_valid_out, data_out, valid_out)
        '''

        if self.config[f"tile_en"] == 0:
            emptyout = [0] * self.interconnect_output_ports
            return emptyout, emptyout, emptyout, emptyout

        (ac_wen_out,
         ac_ren_out,
         ac_valid_out_dat,
         ac_valid_out_stencil) = self.app_ctrl.interact(wen_in=wen,
                                                        ren_in=ren,
                                                        tb_valid=self.tb_valid,
                                                        ren_update=self.tb_valid)

        (arb_wen_en,
         arb_ren_en,
         acc_valid_out_dat,
         acc_valid_out_stencil) = self.app_ctrl_coarse.interact(wen_in=self.iac_port_out & self.valid_agg_buff,
                                                                ren_in=1,
                                                                tb_valid=0,
                                                                ren_update=ack_reduced)

        if self.align_input:
            valids_align = []
            aligns_align = []
            data_align = []
            # Pass data through agg aligners
            for i in range(self.interconnect_input_ports):
                (d_out, v_out, a_out) = self.agg_aligners[i].interact(data_in[i],
                                                                      ac_wen_out[i])
                data_align.append(d_out)
                valids_align.append(v_out)
                aligns_align.append(a_out)
        elif self.agg_height > 0:
            aligns_align = [0] * self.interconnect_input_ports

        if self.agg_height > 0:
            # Now send data to agg buffers
            data_agg_buff = []
            valid_agg_buff = []
            for i in range(self.interconnect_input_ports):
                # (dt, vt) = self.agg_buffs[i].insert(data_align[i], valids_align[i])
                (dt, vt) = self.agg_buffs[i].interact(data_align[i], valids_align[i], 0)
                data_agg_buff.append(dt)
                valid_agg_buff.append(vt)
                # data_agg_buff.append(self.agg_buffs[i].get_item())
                # valid_agg_buff.append(self.agg_buffs[i].get_valid_out())
        elif self.agg_height == 0:
            valid_agg_buff = wen
            data_agg_buff = data_in

        self.valid_agg_buff = valid_agg_buff

        # Now send agg_buff stuff to IAC
        (iac_valid, iac_data, iac_addrs, self.iac_port_out) = \
            self.iac.interact(self.valid_agg_buff, data_agg_buff, arb_wen_en)

        # Get prefetch step for OAC...
        pref_step = []
        arb_ack = 0
        for i in range(self.interconnect_output_ports):
            pref_step.append(self.prefetchers[i].get_step())

        # Rd sync gate for acks from rw arb
        rd_sync_gate = self.sync_groups.get_rd_sync()
        # Also get ren from oac
        oac_ren = self.oac.get_ren(pref_step)

        # OAC
        # First need to get rw_arb acks based on ren + sync group gate
        # Now we need to squash the acks
        ack_base = 0

        for i in range(self.banks):
            ren_local = []
            for j in range(self.interconnect_output_ports):
                ren_local.append(0)
            for j in range(self.interconnect_output_ports):
                ren_local[j] = ren_local[j] | (rd_sync_gate[j] & oac_ren[i][j])
            local_ack = self.rw_arbs[i].get_ack(iac_valid[i], ac_wen_en, ren_local, ac_ren_en)
            ack_base = (ack_base | local_ack)

        # Final interaction with OAC
        (oac_ren, oac_addrs) = self.oac.interact(valid_in=pref_step,
                                                 step_in=ack_base,
                                                 enable_chain_output=self.config[f"enable_chain_output"])

        # Get data from mem
        data_to_arb = []
        for i in range(self.banks):
            if self.read_delay == 1:
                data_to_arb.append(self.mems[i].get_rd_reg())
            else:
                data_to_arb.append(0)

        # RW arbs
        rw_out_dat = []
        rw_out_port = []
        rw_out_valid = []
        rw_cen_mem = []
        rw_wen_mem = []
        rw_data_to_mem = []
        rw_addr_to_mem = []
        rw_ack = []
        rw_mem_valid_data = []
        for i in range(self.banks):
            gated_ren = []
            for j in range(self.interconnect_output_ports):
                gated_ren.append(oac_ren[i][j] & rd_sync_gate[j])

            (od,
             op,
             ov,
             cm,
             wm,
             dm,
             am,
             ack,
             omvd) = self.rw_arbs[i].interact(iac_valid[i],
                                              ac_wen_en,
                                              iac_data[i],
                                              iac_addrs[i],
                                              data_to_arb[i],
                                              gated_ren,
                                              ac_ren_en,
                                              oac_addrs,
                                              self.mem_valid_data)

            rw_out_dat.append(od)
            rw_out_port.append(op)
            rw_out_valid.append(ov)
            rw_cen_mem.append(cm)
            rw_wen_mem.append(wm)
            rw_data_to_mem.append(dm)
            rw_addr_to_mem.append(am)
            rw_ack.append(ack)
            rw_mem_valid_data.append(omvd)

        self.mem_data_out = []
        self.mem_valid_data = []
        # SRAM Wrapper
        if self.read_delay == 1:
            for i in range(self.banks):
                sram_mdo, sram_mvd = self.mems[i].interact(rw_data_to_mem[i],
                                                           rw_addr_to_mem[i],
                                                           rw_cen_mem[i],
                                                           rw_wen_mem[i],
                                                           1,
                                                           1)
                self.mem_data_out.append(sram_mdo)
                self.mem_valid_data.append(sram_mvd)

        # Register File
        else:
            rw_out_dat = []
            for i in range(self.banks):
                local_rf_read = self.mems[i].interact(rw_wen_mem[i],
                                                      rw_addr_to_mem[i],
                                                      rw_addr_to_mem[i],
                                                      rw_data_to_mem[i])
                # Hack for one port
                rw_out_dat.append(local_rf_read[0])

        # Demux those reads
        (demux_dat, demux_valid, demux_mem_valid_data) = \
            self.demux_reads.interact(rw_out_dat,
                                      rw_out_valid,
                                      rw_out_port,
                                      rw_mem_valid_data)

        # Requires or'd version of ren
        ren_base = []
        for j in range(self.interconnect_output_ports):
            ren_base.append(0)
        for j in range(self.interconnect_output_ports):
            for i in range(self.banks):
                ren_base[j] = ren_base[j] | (oac_ren[i][j])

        # Sync groups now
        (sync_data, sync_valid, rd_sync_gate_x, sync_mem_valid_data) = \
            self.sync_groups.interact(ack_base,
                                      demux_dat,
                                      demux_valid,
                                      ren_base,
                                      demux_mem_valid_data)

        # Now get the tba rdy and interact with prefetcher
        tba_rdys = []
        for i in range(self.interconnect_output_ports):
            tba_rdys.append(self.tbas[i].get_ready())

        pref_data = []
        pref_valid = []
        for i in range(self.interconnect_output_ports):
            (pd, pv, psx, prefetcher_mem_valid_data) = \
                self.prefetchers[i].interact(sync_data[i],
                                             sync_valid[i],
                                             tba_rdys[i],
                                             sync_mem_valid_data[i])
            if type(pd) == list:
                pref_data.append(pd.copy())
            else:
                pref_data.append(pd)
            pref_valid.append(pv)

        # Now send this to the TBAs...
        data_out = []
        valid_out = []
        for i in range(self.interconnect_output_ports):
            (tb_d, tb_v) = self.tbas[i].tba_main(pref_data[i],
                                                 pref_valid[i],
                                                 pref_valid[i],
                                                 0,
                                                 ac_ren_out[i],
                                                 prefetcher_mem_valid_data[i])
            data_out.append(tb_d)
            valid_out.append(tb_v)

        self.tb_valid = valid_out

        # Chaining module is completely combinational,
        # just have it here as part of top

        curr_tile_data_out = data_out
        curr_tile_valid_out = valid_out

        chain_data_out_inter = []
        chain_valid_out_inter = []
        for i in range(self.interconnect_output_ports):
            if chain_valid_in[i] == 0:
                chain_data_out_inter.append(curr_tile_data_out[i])
                chain_valid_out_inter.append(curr_tile_valid_out[i])
            else:
                chain_data_out_inter.append(chain_data_in[i])
                chain_valid_out_inter.append(chain_valid_in[i])

        # all combinational outputs
        # set data_out_tile
        if self.enable_chain_output:
            data_out_tile = chain_data_out_inter
        else:
            data_out_tile = curr_tile_data_out

        # set valid_out_tile
        valid_out_tile = []
        if self.enable_chain_output:
            if not (self.chain_idx_output == 0):
                for i in range(self.interconnect_output_ports):
                    valid_out_tile.append(0)
            else:
                valid_out_tile = chain_valid_out_inter
        else:
            valid_out_tile = curr_tile_valid_out

        # set chain_data_out
        chain_data_out = chain_data_out_inter

        # set chain_valid_out
        chain_valid_out = []
        if (self.chain_idx_output == 0) or \
                (not self.enable_chain_output):
            for i in range(self.interconnect_output_ports):
                chain_valid_out.append(0)
        else:
            chain_valid_out = chain_valid_out_inter

        return (chain_data_out, chain_valid_out, data_out_tile, valid_out_tile)



from lake.models.model import Model
from lake.models.reg_fifo_model import RegFIFOModel


class PrefetcherModel(Model):
    def __init__(self,
                 fetch_width,
                 data_width,
                 max_prefetch):

        self.fetch_width = fetch_width
        self.max_prefetch = max_prefetch
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)

        self.config = {}
        self.config['input_latency'] = 0

        self.fifo = RegFIFOModel(data_width=self.data_width,
                                 width_mult=self.fw_int,
                                 depth=self.max_prefetch)

        self.cnt = 0

    def set_config(self, new_config):
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

    def update_cnt(self, valid_read, tba_rdy):
        if valid_read != 0 and tba_rdy == 0:
            self.cnt -= 1
        elif valid_read != 0 & tba_rdy != 0 and not self.fifo.get_full():
            self.cnt += 1

    def get_step(self):
        return int((self.cnt + self.config['input_latency']) < (self.max_prefetch))

    def get_cnt(self):
        return self.cnt

    def interact(self, data_in, valid_read, tba_rdy, mem_valid_data):
        '''
        Returns (data, valid, step, mem_valid_data_out)
        '''
        (d_out, v_out, empty, full, mem_valid_data_out) = \
            self.fifo.interact(valid_read, tba_rdy, data_in, mem_valid_data)
        stp = self.get_step()
        self.update_cnt(valid_read, tba_rdy)
        if type(d_out) == list:
            return (d_out.copy(), v_out, stp, mem_valid_data_out)
        else:
            return (d_out, v_out, stp, mem_valid_data_out)



from kratos import *
from math import log


class Aggregator(Generator):
    def __init__(self,
                 word_width,
                 mem_word_width):

        super().__init__("aggregator", debug=True)
        self.mem_word_width = mem_word_width
        # inputs
        self.clk = self.clock("clk")
        # active low asynchornous reset
        self.rst_n = self.reset("rst_n", 1)
        self.in_pixels = self.input("in_pixels",
                                    word_width)
        self._valid_in = self.input("valid_in", 1)
        self._align = self.input("align", 1)

        # outputs
        self._agg_out = self.output("agg_out",
                                    word_width,
                                    size=mem_word_width,
                                    packed=True,
                                    explicit_array=True)
        self._valid_out = self.output("valid_out", 1)
        self._next_full = self.output("next_full", 1)

        self.shift_reg = self.var("shift_reg",
                                  width=word_width,
                                  size=mem_word_width,
                                  explicit_array=True,
                                  packed=True)

        # local variables
        if(mem_word_width > 1):
            self._word_count = self.var("word_count", clog2(mem_word_width))
            self.add_code(self.update_counter_valid)
            self.add_code(self.set_next_full)
            self.add_code(self.update_shift_reg)
            # add combinational blocks
        else:
            self.wire(self._valid_out, const(1, 1))
            self.wire(self._next_full, self._valid_in)
            # only add the update counter/valid
            self.wire(self.shift_reg[0], self.in_pixels)

        self.add_code(self.output_data)

    @always_comb
    def set_next_full(self):
        self._next_full = (self._valid_in & (self._word_count == const(self.mem_word_width - 1,
                                                                       self._word_count.width))) | self._align

    # setting valid signal and word_count index
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_counter_valid(self):
        if ~self.rst_n:
            self._valid_out = 0
            self._word_count = const(0, self._word_count.width)
        # no self._word_count in this case
        elif self._valid_in:
            if (self._word_count == const(self.mem_word_width - 1, self._word_count.width)) | self._align:
                self._valid_out = const(1, 1)
                self._word_count = const(0, self._word_count.width)
            else:
                self._valid_out = const(0, 1)
                self._word_count = self._word_count + const(1, clog2(mem_word_width))
        else:
            self._valid_out = 0

    @always_ff((posedge, "clk"))
    def update_shift_reg(self):
        if (self._valid_in):
            self.shift_reg[self._word_count] = self.in_pixels

    @always_comb
    def output_data(self):
        self._agg_out = self.shift_reg


if __name__ == "__main__":
    db_dut = Aggregator(word_width=16,
                        mem_word_width=4)
    verilog(db_dut, filename="aggregator.sv")



import kratos as kts
from kratos import *
from lake.passes.passes import lift_config_reg
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.utils.util import extract_formal_annotation, safe_wire, add_counter, trim_config_list
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.control_signal_attr import ControlSignalAttr
from lake.modules.storage_config_seq import StorageConfigSeq
from lake.utils.parse_clkwork_config import extract_controller_json, map_controller
from lake.utils.parse_clkwork_config import ControllerInfo
from _kratos import create_wrapper_flatten


class Pond(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_depth=32,
                 default_iterator_support=2,
                 interconnect_input_ports=1,  # Connection to int
                 interconnect_output_ports=1,
                 config_data_width=32,
                 config_addr_width=8,
                 cycle_count_width=16,
                 add_clk_enable=True,
                 add_flush=True):
        super().__init__("pond", debug=True)

        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.mem_depth = mem_depth
        self.data_width = data_width
        self.config_data_width = config_data_width
        self.config_addr_width = config_addr_width
        self.add_clk_enable = add_clk_enable
        self.add_flush = add_flush
        self.cycle_count_width = cycle_count_width
        self.default_iterator_support = default_iterator_support
        self.default_config_width = kts.clog2(self.mem_depth)
        # inputs
        self._clk = self.clock("clk")
        self._clk.add_attribute(FormalAttr(f"{self._clk.name}", FormalSignalConstraint.CLK))
        self._rst_n = self.reset("rst_n")
        self._rst_n.add_attribute(FormalAttr(f"{self._rst_n.name}", FormalSignalConstraint.RSTN))
        self._clk_en = self.clock_en("clk_en", 1)

        # Enable/Disable tile
        self._tile_en = self.input("tile_en", 1)
        self._tile_en.add_attribute(ConfigRegAttr("Tile logic enable manifested as clock gate"))

        gclk = self.var("gclk", 1)
        self._gclk = kts.util.clock(gclk)
        self.wire(gclk, kts.util.clock(self._clk & self._tile_en))

        self._cycle_count = add_counter(self, "cycle_count", self.cycle_count_width)

        # Create write enable + addr, same for read.
        # self._write = self.input("write", self.interconnect_input_ports)
        self._write = self.var("write", self.interconnect_input_ports)
        # self._write.add_attribute(ControlSignalAttr(is_control=True))

        self._write_addr = self.var("write_addr",
                                    kts.clog2(self.mem_depth),
                                    size=self.interconnect_input_ports,
                                    explicit_array=True,
                                    packed=True)

        # Add "_pond" suffix to avoid error during garnet RTL generation
        self._data_in = self.input("data_in_pond", self.data_width,
                                   size=self.interconnect_input_ports,
                                   explicit_array=True,
                                   packed=True)
        self._data_in.add_attribute(FormalAttr(f"{self._data_in.name}", FormalSignalConstraint.SEQUENCE))
        self._data_in.add_attribute(ControlSignalAttr(is_control=False))

        self._read = self.var("read", self.interconnect_output_ports)
        # self._read.add_attribute(ControlSignalAttr(is_control=True))

        self._read_addr = self.var("read_addr",
                                   kts.clog2(self.mem_depth),
                                   size=self.interconnect_output_ports,
                                   explicit_array=True,
                                   packed=True)
        self._data_out = self.output("data_out_pond", self.data_width,
                                     size=self.interconnect_output_ports,
                                     explicit_array=True,
                                     packed=True)
        self._data_out.add_attribute(FormalAttr(f"{self._data_out.name}", FormalSignalConstraint.SEQUENCE))
        self._data_out.add_attribute(ControlSignalAttr(is_control=False))

        self._valid_out = self.output("valid_out_pond", self.interconnect_output_ports)
        self._valid_out.add_attribute(FormalAttr(f"{self._valid_out.name}", FormalSignalConstraint.SEQUENCE))
        self._valid_out.add_attribute(ControlSignalAttr(is_control=False))

        self._mem_data_out = self.var("mem_data_out", self.data_width,
                                      size=self.interconnect_output_ports,
                                      explicit_array=True,
                                      packed=True)

        self._mem_data_in = self.var("mem_data_in", self.data_width,
                                     size=self.interconnect_input_ports,
                                     explicit_array=True,
                                     packed=True)

        self._mem_write_addr = self.var("mem_write_addr",
                                        kts.clog2(self.mem_depth),
                                        size=self.interconnect_input_ports,
                                        explicit_array=True,
                                        packed=True)

        self._mem_read_addr = self.var("mem_read_addr",
                                       kts.clog2(self.mem_depth),
                                       size=self.interconnect_output_ports,
                                       explicit_array=True,
                                       packed=True)

        self.wire(self._data_out, self._mem_data_out)

        # Valid out is simply passing the read signal through...
        self.wire(self._valid_out, self._read)

        # Create write addressors
        for wr_port in range(self.interconnect_input_ports):

            RF_WRITE_ITER = ForLoop(iterator_support=self.default_iterator_support,
                                    config_width=self.cycle_count_width)
            RF_WRITE_ADDR = AddrGen(iterator_support=self.default_iterator_support,
                                    config_width=self.default_config_width)
            RF_WRITE_SCHED = SchedGen(iterator_support=self.default_iterator_support,
                                      config_width=self.cycle_count_width,
                                      use_enable=True)

            self.add_child(f"rf_write_iter_{wr_port}",
                           RF_WRITE_ITER,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           step=self._write[wr_port])
            # Whatever comes through here should hopefully just pipe through seamlessly
            # addressor modules
            self.add_child(f"rf_write_addr_{wr_port}",
                           RF_WRITE_ADDR,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           step=self._write[wr_port],
                           mux_sel=RF_WRITE_ITER.ports.mux_sel_out,
                           restart=RF_WRITE_ITER.ports.restart)
            safe_wire(self, self._write_addr[wr_port], RF_WRITE_ADDR.ports.addr_out)

            self.add_child(f"rf_write_sched_{wr_port}",
                           RF_WRITE_SCHED,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           mux_sel=RF_WRITE_ITER.ports.mux_sel_out,
                           finished=RF_WRITE_ITER.ports.restart,
                           cycle_count=self._cycle_count,
                           valid_output=self._write[wr_port])

        # Create read addressors
        for rd_port in range(self.interconnect_output_ports):

            RF_READ_ITER = ForLoop(iterator_support=self.default_iterator_support,
                                   config_width=self.cycle_count_width)
            RF_READ_ADDR = AddrGen(iterator_support=self.default_iterator_support,
                                   config_width=self.default_config_width)
            RF_READ_SCHED = SchedGen(iterator_support=self.default_iterator_support,
                                     config_width=self.cycle_count_width,
                                     use_enable=True)

            self.add_child(f"rf_read_iter_{rd_port}",
                           RF_READ_ITER,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           step=self._read[rd_port])

            self.add_child(f"rf_read_addr_{rd_port}",
                           RF_READ_ADDR,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           step=self._read[rd_port],
                           mux_sel=RF_READ_ITER.ports.mux_sel_out,
                           restart=RF_READ_ITER.ports.restart)
            safe_wire(self, self._read_addr[rd_port], RF_READ_ADDR.ports.addr_out)

            self.add_child(f"rf_read_sched_{rd_port}",
                           RF_READ_SCHED,
                           clk=self._gclk,
                           rst_n=self._rst_n,
                           mux_sel=RF_READ_ITER.ports.mux_sel_out,
                           finished=RF_READ_ITER.ports.restart,
                           cycle_count=self._cycle_count,
                           valid_output=self._read[rd_port])

        # ===================================
        # Instantiate config hooks...
        # ===================================
        self.fw_int = 1
        self.data_words_per_set = 2 ** self.config_addr_width
        self.sets = int((self.fw_int * self.mem_depth) / self.data_words_per_set)

        self.sets_per_macro = max(1, int(self.mem_depth / self.data_words_per_set))
        self.total_sets = max(1, 1 * self.sets_per_macro)

        self._config_data_in = self.input("config_data_in",
                                          self.config_data_width)
        self._config_data_in.add_attribute(ControlSignalAttr(is_control=False))

        self._config_data_in_shrt = self.var("config_data_in_shrt",
                                             self.data_width)

        self.wire(self._config_data_in_shrt, self._config_data_in[self.data_width - 1, 0])

        self._config_addr_in = self.input("config_addr_in",
                                          self.config_addr_width)
        self._config_addr_in.add_attribute(ControlSignalAttr(is_control=False))

        self._config_data_out_shrt = self.var("config_data_out_shrt", self.data_width,
                                              size=self.total_sets,
                                              explicit_array=True,
                                              packed=True)

        self._config_data_out = self.output("config_data_out", self.config_data_width,
                                            size=self.total_sets,
                                            explicit_array=True,
                                            packed=True)
        self._config_data_out.add_attribute(ControlSignalAttr(is_control=False))

        for i in range(self.total_sets):
            self.wire(self._config_data_out[i],
                      self._config_data_out_shrt[i].extend(self.config_data_width))

        self._config_read = self.input("config_read", 1)
        self._config_read.add_attribute(ControlSignalAttr(is_control=False))

        self._config_write = self.input("config_write", 1)
        self._config_write.add_attribute(ControlSignalAttr(is_control=False))

        self._config_en = self.input("config_en", self.total_sets)
        self._config_en.add_attribute(ControlSignalAttr(is_control=False))

        self._mem_data_cfg = self.var("mem_data_cfg", self.data_width,
                                      explicit_array=True,
                                      packed=True)

        self._mem_addr_cfg = self.var("mem_addr_cfg", kts.clog2(self.mem_depth))

        # Add config...
        stg_cfg_seq = StorageConfigSeq(data_width=self.data_width,
                                       config_addr_width=self.config_addr_width,
                                       addr_width=kts.clog2(self.mem_depth),
                                       fetch_width=self.data_width,
                                       total_sets=self.total_sets,
                                       sets_per_macro=self.sets_per_macro)

        # The clock to config sequencer needs to be the normal clock or
        # if the tile is off, we bring the clock back in based on config_en
        cfg_seq_clk = self.var("cfg_seq_clk", 1)
        self._cfg_seq_clk = kts.util.clock(cfg_seq_clk)
        self.wire(cfg_seq_clk, kts.util.clock(self._gclk))

        self.add_child(f"config_seq", stg_cfg_seq,
                       clk=self._cfg_seq_clk,
                       rst_n=self._rst_n,
                       clk_en=self._clk_en | self._config_en.r_or(),
                       config_data_in=self._config_data_in_shrt,
                       config_addr_in=self._config_addr_in,
                       config_wr=self._config_write,
                       config_rd=self._config_read,
                       config_en=self._config_en,
                       wr_data=self._mem_data_cfg,
                       rd_data_out=self._config_data_out_shrt,
                       addr_out=self._mem_addr_cfg)

        if self.interconnect_output_ports == 1:
            self.wire(stg_cfg_seq.ports.rd_data_stg, self._mem_data_out)
        else:
            self.wire(stg_cfg_seq.ports.rd_data_stg[0], self._mem_data_out[0])

        self.RF_GEN = RegisterFile(data_width=self.data_width,
                                   write_ports=self.interconnect_input_ports,
                                   read_ports=self.interconnect_output_ports,
                                   width_mult=1,
                                   depth=self.mem_depth,
                                   read_delay=0)

        # Now we can instantiate and wire up the register file
        self.add_child(f"rf",
                       self.RF_GEN,
                       clk=self._gclk,
                       rst_n=self._rst_n,
                       data_out=self._mem_data_out)

        # Opt in for config_write
        self._write_rf = self.var("write_rf", self.interconnect_input_ports)
        self.wire(self._write_rf[0], kts.ternary(self._config_en.r_or(), self._config_write, self._write[0]))
        for i in range(self.interconnect_input_ports - 1):
            self.wire(self._write_rf[i + 1], kts.ternary(self._config_en.r_or(), kts.const(0, 1), self._write[i + 1]))
        self.wire(self.RF_GEN.ports.wen, self._write_rf)

        # Opt in for config_data_in
        self.wire(self._mem_data_in[0], kts.ternary(self._config_en.r_or(), self._mem_data_cfg, self._data_in[0]))
        for i in range(self.interconnect_input_ports - 1):
            self.wire(self._mem_data_in[i + 1], self._data_in[i + 1])
        self.wire(self.RF_GEN.ports.data_in, self._mem_data_in)

        # Opt in for config_addr
        self.wire(self._mem_write_addr[0], kts.ternary(self._config_en.r_or(),
                                                       self._mem_addr_cfg,
                                                       self._write_addr[0]))
        for i in range(self.interconnect_input_ports - 1):
            self.wire(self._mem_write_addr[i + 1], self._write_addr[i + 1])
        if self.interconnect_input_ports == 1:
            self.wire(self.RF_GEN.ports.wr_addr, self._mem_write_addr[0])
        else:
            self.wire(self.RF_GEN.ports.wr_addr, self._mem_write_addr)

        self.wire(self._mem_read_addr[0], kts.ternary(self._config_en.r_or(),
                                                      self._mem_addr_cfg,
                                                      self._read_addr[0]))
        for i in range(self.interconnect_output_ports - 1):
            self.wire(self._mem_read_addr[i + 1], self._read_addr[i + 1])
        if self.interconnect_output_ports == 1:
            self.wire(self.RF_GEN.ports.rd_addr, self._mem_read_addr[0])
        else:
            self.wire(self.RF_GEN.ports.rd_addr, self._mem_read_addr)

        if self.add_clk_enable:
            # self.clock_en("clk_en")
            kts.passes.auto_insert_clock_enable(self.internal_generator)
            clk_en_port = self.internal_generator.get_port("clk_en")
            clk_en_port.add_attribute(ControlSignalAttr(False))

        if self.add_flush:
            self.add_attribute("sync-reset=flush")
            kts.passes.auto_insert_sync_reset(self.internal_generator)
            flush_port = self.internal_generator.get_port("flush")
            flush_port.add_attribute(ControlSignalAttr(True))

        # Finally, lift the config regs...
        lift_config_reg(self.internal_generator)

    def get_static_bitstream_json(self,
                                  root_node):

        # Dummy variables to fill in later when compiler
        # generates different collateral for different designs
        flattened = create_wrapper_flatten(self.internal_generator.clone(),
                                           self.name + "_W")

        # Store all configurations here
        config = []

        # Get controllers from json node...
        assert "in2regfile" in root_node
        assert "regfile2out" in root_node
        in2rf_ctrl = map_controller(extract_controller_json(root_node["in2regfile"]), "in2regfile")
        rf2out_ctrl = map_controller(extract_controller_json(root_node["regfile2out"]), "regfile2out")

        # Configure registers based on controller data...
        config.append(("rf_write_iter_0_dimensionality", in2rf_ctrl.dim))
        config.append(("rf_write_addr_0_starting_addr", in2rf_ctrl.in_data_strt))
        config.append(("rf_write_sched_0_sched_addr_gen_starting_addr", in2rf_ctrl.cyc_strt))
        config.append(("rf_write_sched_0_enable", 1))
        for i in range(in2rf_ctrl.dim):
            config.append((f"rf_write_addr_0_strides_{i}", in2rf_ctrl.in_data_stride[i]))
            config.append((f"rf_write_iter_0_ranges_{i}", in2rf_ctrl.extent[i]))
            config.append((f"rf_write_sched_0_sched_addr_gen_strides_{i}", in2rf_ctrl.cyc_stride[i]))

        config.append(("rf_read_iter_0_dimensionality", rf2out_ctrl.dim))
        config.append(("rf_read_addr_0_starting_addr", rf2out_ctrl.out_data_strt))
        config.append(("rf_read_sched_0_sched_addr_gen_starting_addr", rf2out_ctrl.cyc_strt))
        config.append(("rf_read_sched_0_enable", 1))

        for i in range(rf2out_ctrl.dim):
            config.append((f"rf_read_addr_0_strides_{i}", rf2out_ctrl.out_data_stride[i]))
            config.append((f"rf_read_iter_0_ranges_{i}", rf2out_ctrl.extent[i]))
            config.append((f"rf_read_sched_0_sched_addr_gen_strides_{i}", rf2out_ctrl.cyc_stride[i]))

        # Handle control registers... (should really be done in garnet TODO)
        config.append(("flush_reg_sel", 0))  # 1
        config.append(("flush_reg_value", 0))  # 1
        # Activate the tile...
        config.append(("tile_en", 1))  # 1

        # Trim the list
        return trim_config_list(flattened, config)


if __name__ == "__main__":
    pond_dut = Pond(data_width=16,  # CGRA Params
                    mem_depth=32,
                    default_iterator_support=2,
                    interconnect_input_ports=1,  # Connection to int
                    interconnect_output_ports=1,
                    cycle_count_width=16,
                    add_clk_enable=True,
                    add_flush=True)

    # Lift config regs and generate annotation
    # lift_config_reg(pond_dut.internal_generator)
    extract_formal_annotation(pond_dut, "pond.txt")

    verilog(pond_dut, filename="pond.sv",
            optimize_if=False)



from kratos import *
from lake.utils.util import increment, decrement
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.modules.reg_fifo import RegFIFO


class Prefetcher(Generator):
    '''
    This module will handle prefetching from SRAM to the tb.
    Since there is a variable latency for a ready cycle to
    provide the next valid set of data to the tbs, we need to
    buffer some amount so that we can satisfy these requests with immediacy
    '''
    def __init__(self,
                 fetch_width,
                 data_width,
                 max_prefetch):
        super().__init__("prefetcher")
        # Capture to the object
        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.max_prefetch = max_prefetch

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=self.fw_int,
                                   explicit_array=True,
                                   packed=True)

        self._valid_read = self.input("valid_read", 1)
        self._tba_rdy_in = self.input("tba_rdy_in", 1)

        self._input_latency = self.input("input_latency", clog2(self.max_prefetch) + 1)
        doc = "This register is set to denote the input latency loop for reads. " + \
              "This is sent to an internal fifo and an almost full signal is " + \
              "used to pull more reads that the transpose buffers need."
        self._input_latency.add_attribute(ConfigRegAttr(doc))

        self._max_lat = const(self.max_prefetch - 1,
                              clog2(self.max_prefetch) + 1)

        # Outputs
        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=self.fw_int,
                                     explicit_array=True,
                                     packed=True)

        self._valid_out = self.output("valid_out", 1)
        self._prefetch_step = self.output("prefetch_step", 1)

        # Local Signals
        self._cnt = self.var("cnt", clog2(self.max_prefetch) + 1)
        self._fifo_empty = self.var("fifo_empty", 1)
        self._fifo_full = self.var("fifo_full", 1)

        reg_fifo = RegFIFO(data_width=self.data_width,
                           width_mult=self.fw_int,
                           depth=self.max_prefetch)

        self.add_child("fifo", reg_fifo,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       clk_en=1,
                       data_in=self._data_in,
                       data_out=self._data_out,
                       push=self._valid_read,
                       pop=self._tba_rdy_in,
                       empty=self._fifo_empty,
                       full=self._fifo_full,
                       valid=self._valid_out)

        # Generate
        self.add_code(self.update_cnt)
        self.add_code(self.set_prefetch_step)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_cnt(self):
        if ~self._rst_n:
            self._cnt = 0
        elif(self._valid_read & ~self._tba_rdy_in & ~self._fifo_full):
            self._cnt = self._cnt + 1
        elif(~self._valid_read & self._tba_rdy_in & ~self._fifo_empty):
            self._cnt = self._cnt - 1

    @always_comb
    def set_prefetch_step(self):
        self._prefetch_step = (self._cnt + self._input_latency) < (self.max_prefetch)


if __name__ == "__main__":
    align_dut = Prefetcher(fetch_width=32,
                           data_width=16,
                           max_prefetch=64)
    verilog(align_dut, filename="prefetcher.sv")



from lake.models.model import Model
from lake.models.addr_gen_model import AddrGenModel
import math as mt
import kratos as kts


class OutputAddrCtrlModel(Model):

    def __init__(self,
                 interconnect_output_ports,
                 mem_depth,
                 num_tiles,
                 banks,
                 iterator_support,
                 address_width,
                 data_width,
                 fetch_width,
                 chain_idx_output):

        self.interconnect_output_ports = interconnect_output_ports
        self.mem_depth = mem_depth
        self.num_tiles = num_tiles
        self.banks = banks
        self.iterator_support = iterator_support
        self.address_width = address_width
        self.data_width = data_width
        self.fetch_width = fetch_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.chain_idx_output = chain_idx_output

        self.config = {}

        # Create child address generators
        self.addr_gens = []
        for i in range(self.interconnect_output_ports):
            new_addr_gen = AddrGenModel(iterator_support=self.iterator_support,
                                        address_width=self.address_width)
            self.addr_gens.append(new_addr_gen)

        self.mem_addr_width = kts.clog2(self.num_tiles * self.mem_depth)
        self.chain_idx_bits = max(1, kts.clog2(self.num_tiles))

        # Get local list of addresses
        self.addresses = []
        for i in range(self.interconnect_output_ports):
            self.addresses.append(0)

        # Initialize the configuration
        for i in range(self.interconnect_output_ports):
            self.config[f"address_gen_{i}_starting_addr"] = 0
            self.config[f"address_gen_{i}_dimensionality"] = 0
            for j in range(self.iterator_support):
                self.config[f"address_gen_{i}_strides_{j}"] = 0
                self.config[f"address_gen_{i}_ranges_{j}"] = 0

        # Set up the wen
        self.ren = []
        self.mem_addresses = []
        for i in range(self.banks):
            self.ren.append([])
            for j in range(self.interconnect_output_ports):
                self.ren[i].append(0)
        for i in range(self.interconnect_output_ports):
            self.mem_addresses.append(0)

    def set_config(self, new_config):
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val
        # Configure children
        for i in range(self.interconnect_output_ports):
            addr_gen_config = {}
            addr_gen_config["starting_addr"] = self.config[f"address_gen_{i}_starting_addr"]
            addr_gen_config["dimensionality"] = self.config[f"address_gen_{i}_dimensionality"]
            for j in range(self.iterator_support):
                addr_gen_config[f"strides_{j}"] = self.config[f"address_gen_{i}_strides_{j}"]
                addr_gen_config[f"ranges_{j}"] = self.config[f"address_gen_{i}_ranges_{j}"]
            self.addr_gens[i].set_config(addr_gen_config)

    def interact(self, valid_in, step_in, enable_chain_output):
        '''
        Returns (ren, addrs)
        '''
        ren = self.get_ren(valid_in)
        addrs = self.get_addrs_tile_en()
        self.step_addrs(valid_in, step_in)
        return (ren, addrs)

    # Retrieve the current addresses from each generator
    def get_addrs_tile_en(self):
        for i in range(self.interconnect_output_ports):
            to_get = self.addr_gens[i]
            self.addresses[i] = to_get.get_address() % self.mem_depth
            addr_chain_bits = (self.addresses[i]) >> (self.mem_addr_width - self.chain_idx_bits - 1)
        return self.addresses

    def get_addrs_full(self):
        for i in range(self.interconnect_output_ports):
            to_get = self.addr_gens[i]
            self.addresses[i] = to_get.get_address()
        return self.addresses

    # Get the ren for the current valid input
    def get_ren(self, valid):
        for i in range(self.banks):
            for j in range(self.interconnect_output_ports):
                self.ren[i][j] = 0
        for i in range(self.interconnect_output_ports):
            if(valid[i]):
                if(self.banks == 1):
                    self.ren[0][i] = 1
                else:
                    self.ren[self.get_addrs_full()[i] >> (self.mem_addr_width)][i] = 1
        return self.ren

    # Step the addresses based on valid
    def step_addrs(self, valid, step):
        for i, valid_input in enumerate(valid):
            if valid_input & ((step & (1 << i)) != 0):
                to_step = self.addr_gens[i]
                to_step.step()

        # Not implemented
    def update_ports(self):
        raise NotImplementedError

    def peek(self):
        raise NotImplementedError



import kratos as kts


# this attribute indicates sram ports which may be renamed if macro being used requires
# different port names
class SRAMPortAttr(kts.Attribute):
    def __init__(self,
                 doc_string=""):
        super().__init__()
        self.value = "sram_port"
        self.documentation = doc_string

    def set_documentation(self, new_doc):
        self.documentation = new_doc

    def get_documentation(self):
        return self.documentation



from lake.models.model import Model


class RWArbiterModel(Model):
    def __init__(self,
                 fetch_width,
                 data_width,
                 memory_depth,
                 int_out_ports,
                 read_delay):
        self.fetch_width = fetch_width
        self.memory_depth = memory_depth
        self.int_out_ports = int_out_ports
        self.read_delay = read_delay

        self.rd_valid = 0
        self.rd_port = 0
        self.ack = 0

    def set_config(self, new_config):
        # No configuration space
        return

    def interact(self,
                 wen_in,
                 wen_en,
                 w_data,
                 w_addr,
                 data_from_mem,
                 ren_in,
                 ren_en,
                 rd_addr,
                 mem_valid_data):
        '''
        Returns (out_dat, out_port, out_valid,
                 cen_mem, wen_mem, data_to_mem, addr_to_mem, ack,
                 out_mem_valid_data)
        '''
        # These signals are always this way
        out_dat = data_from_mem
        if self.read_delay == 1:
            out_port = self.rd_port
            out_valid = self.rd_valid
        else:
            out_port = 0
            out_valid = 0
        data_to_mem = w_data
        if type(wen_en) is list:
            wen_en = wen_en[0]
        wen_mem = wen_in & wen_en
        # Signals following may vary
        cen_mem = 0
        addr_to_mem = 0

        self.rd_valid = 0
        if wen_in != 0 and wen_en != 0:
            cen_mem = 1
            addr_to_mem = w_addr
            self.rd_valid = 0
            self.rd_port = 0
        elif ren_in != 0 and ren_en != 0:
            for i in range(self.int_out_ports):
                # Select lowest
                if ren_in[i] != 0 and ren_en[i] != 0:
                    cen_mem = 1
                    addr_to_mem = rd_addr[i]
                    self.rd_valid = 1
                    self.rd_port = 1 << i
                    if self.read_delay == 0:
                        out_valid = 1
                        out_port = 1 << i
                    break
        ack = self.get_ack(wen_in, wen_en, ren_in, ren_en)

        out_mem_valid_data = mem_valid_data

        return (out_dat, out_port, out_valid,
                cen_mem, wen_mem, data_to_mem,
                addr_to_mem, ack, out_mem_valid_data)

    def get_ack(self, wen_in, wen_en, ren_in, ren_en):
        self.ack = 0
        if wen_in != 0 and wen_en != 0:
            self.ack = 0
        elif ren_in != 0 and ren_en != 0:
            for i in range(self.int_out_ports):
                # Select lowest
                if ren_in[i] != 0 and ren_en[i]:
                    self.ack = 1 << i
                    break
        return self.ack
